"""
Halo callback object



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np
import os

from yt.analysis_modules.cosmological_observation.light_ray.light_ray import \
    periodic_distance
from yt.data_objects.profiles import \
    create_profile
from yt.frontends.ytdata.utilities import \
    _hdf5_yt_array, \
    _yt_array_hdf5
from yt.units.yt_array import \
    YTArray
from yt.utilities.exceptions import \
    YTSphereTooSmall
from yt.funcs import \
    ensure_list
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.operator_registry import \
    OperatorRegistry
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_root_only
from yt.visualization.profile_plotter import \
    PhasePlot

callback_registry = OperatorRegistry()
    
def add_callback(name, function):
    callback_registry[name] =  HaloCallback(function)

class HaloCallback(object):
    r"""
    A HaloCallback is a function that minimally takes in a Halo object 
    and performs some analysis on it.  This function may attach attributes 
    to the Halo object, write out data, etc, but does not return anything.
    """
    def __init__(self, function, args=None, kwargs=None):
        self.function = function
        self.args = args
        if self.args is None: self.args = []
        self.kwargs = kwargs
        if self.kwargs is None: self.kwargs = {}

    def __call__(self, halo):
        self.function(halo, *self.args, **self.kwargs)
        return True

def halo_sphere(halo, radius_field="virial_radius", factor=1.0, 
                field_parameters=None):
    r"""
    Create a sphere data container to associate with a halo.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    radius_field : string
        Field to be retrieved from the quantities dictionary as 
        the basis of the halo radius.
        Default: "virial_radius".
    factor : float
        Factor to be multiplied by the base radius for defining 
        the radius of the sphere.
        Default: 1.0.
    field_parameters : dict
        Dictionary of field parameters to be set with the sphere 
        created.
        
    """

    dds = halo.halo_catalog.data_ds
    center = dds.arr([halo.quantities["particle_position_%s" % axis] \
                      for axis in "xyz"])
    radius = factor * halo.quantities[radius_field]
    if radius <= 0.0:
        halo.data_object = None
        return
    try:
        sphere = dds.sphere(center, radius)
    except YTSphereTooSmall:
        halo.data_object = None
        return
    if field_parameters is not None:
        for field, par in field_parameters.items():
            if isinstance(par, tuple) and par[0] == "quantity":
                value = halo.quantities[par[1]]
            else:
                value = par
            sphere.set_field_parameter(field, value)
    halo.data_object = sphere

add_callback("sphere", halo_sphere)

def sphere_field_max_recenter(halo, field):
    r"""
    Recenter the halo sphere on the location of the maximum of the given field.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    field : string
        Field to be used for recentering.

    """

    if halo.data_object is None: return
    s_ds = halo.data_object.ds
    old_sphere = halo.data_object
    max_vals = old_sphere.quantities.max_location(field)
    new_center = s_ds.arr(max_vals[1:])
    new_sphere = s_ds.sphere(new_center.in_units("code_length"),
                               old_sphere.radius.in_units("code_length"))
    mylog.info("Moving sphere center from %s to %s." % (old_sphere.center,
                                                        new_sphere.center))
    for par, value in old_sphere.field_parameters.items():
        if par not in new_sphere.field_parameters:
            new_sphere.set_field_parameter(par, value)
    halo.data_object = new_sphere

add_callback("sphere_field_max_recenter", sphere_field_max_recenter)

def sphere_bulk_velocity(halo):
    r"""
    Set the bulk velocity for the sphere.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.

    """

    halo.data_object.set_field_parameter("bulk_velocity",
        halo.data_object.quantities.bulk_velocity(use_particles=True))

add_callback("sphere_bulk_velocity", sphere_bulk_velocity)

def profile(halo, bin_fields, profile_fields, n_bins=32, extrema=None, logs=None, units=None,
            weight_field="cell_mass", accumulation=False, fractional=False,
            storage="profiles", output_dir="."):
    r"""
    Create 1, 2, or 3D profiles of a halo.

    Store profile data in a dictionary associated with the halo object.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    bin_fields : list of strings
        The binning fields for the profile.
    profile_fields : string or list of strings
        The fields to be profiled.
    n_bins : int or list of ints
        The number of bins in each dimension.  If None, 32 bins for
        each bin are used for each bin field.
        Default: 32.
    extrema : dict of min, max tuples
        Minimum and maximum values of the bin_fields for the profiles.
        The keys correspond to the field names. Defaults to the extrema
        of the bin_fields of the dataset. If a units dict is provided, extrema
        are understood to be in the units specified in the dictionary.
    logs : dict of boolean values
        Whether or not to log the bin_fields for the profiles.
        The keys correspond to the field names. Defaults to the take_log
        attribute of the field.
    units : dict of strings
        The units of the fields in the profiles, including the bin_fields.
    weight_field : string
        Weight field for profiling.
        Default : "cell_mass"
    accumulation : bool or list of bools
        If True, the profile values for a bin n are the cumulative sum of
        all the values from bin 0 to n.  If -True, the sum is reversed so
        that the value for bin n is the cumulative sum from bin N (total bins)
        to n.  If the profile is 2D or 3D, a list of values can be given to
        control the summation in each dimension independently.
        Default: False.
    fractional : If True the profile values are divided by the sum of all
        the profile data such that the profile represents a probability
        distribution function.
    storage : string
        Name of the dictionary to store profiles.
        Default: "profiles"
    output_dir : string
        Name of directory where profile data will be written.  The full path will be
        the output_dir of the halo catalog concatenated with this directory.
        Default : "."

    """

    mylog.info("Calculating 1D profile for halo %d." % 
               halo.quantities["particle_identifier"])

    dds = halo.halo_catalog.data_ds

    if dds is None:
        raise RuntimeError("Profile callback requires a data ds.")

    if not hasattr(halo, "data_object"):
        raise RuntimeError("Profile callback requires a data container.")

    if halo.data_object is None:
        mylog.info("Skipping halo %d since data_object is None." %
                   halo.quantities["particle_identifier"])
        return

    if output_dir is None:
        output_dir = storage
    output_dir = os.path.join(halo.halo_catalog.output_dir, output_dir)
    
    bin_fields = ensure_list(bin_fields)
    my_profile = create_profile(halo.data_object, bin_fields, profile_fields, n_bins=n_bins,
                                extrema=extrema, logs=logs, units=units, weight_field=weight_field,
                                accumulation=accumulation, fractional=fractional)
                  
    prof_store = dict([(field, my_profile[field]) \
                       for field in my_profile.field_data])
    prof_store[my_profile.x_field] = my_profile.x
    if len(bin_fields) > 1:
        prof_store[my_profile.y_field] = my_profile.y
    if len(bin_fields) > 2:
        prof_store[my_profile.z_field] = my_profile.z
    if hasattr(halo, storage):
        halo_store = getattr(halo, storage)
        if "used" in halo_store:
            halo_store["used"] &= my_profile.used
    else:
        halo_store = {"used": my_profile.used}
        setattr(halo, storage, halo_store)
    halo_store.update(prof_store)

    if my_profile.standard_deviation is not None:
        variance_store = dict([(field, my_profile.standard_deviation[field]) \
                               for field in my_profile.standard_deviation])
        variance_storage = "%s_variance" % storage
        if hasattr(halo, variance_storage):
            halo_variance_store = getattr(halo, variance_storage)
        else:
            halo_variance_store = {}
            setattr(halo, variance_storage, halo_variance_store)
        halo_variance_store.update(variance_store)

add_callback("profile", profile)

@parallel_root_only
def save_profiles(halo, storage="profiles", filename=None,
                  output_dir="."):
    r"""
    Save profile data to disk.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    storage : string
        Name of the dictionary attribute containing the profile data to be written.
        Default: "profiles"
    filename : string
        The name of the file to be written.  The final filename will be 
        "<filename>_<id>.h5".  If None, filename is set to the value given 
        by the storage keyword.
        Default: None
    output_dir : string
        Name of directory where profile data will be written.  The full path will be
        the output_dir of the halo catalog concatenated with this directory.
        Default : "."
    
    """

    if not hasattr(halo, storage):
        return
    
    if filename is None:
        filename = storage
    output_file = os.path.join(halo.halo_catalog.output_dir, output_dir,
                               "%s_%06d.h5" % (filename, 
                                               halo.quantities["particle_identifier"]))
    mylog.info("Saving halo %d profile data to %s." %
               (halo.quantities["particle_identifier"], output_file))

    fh = h5py.File(output_file, "w")
    my_profile = getattr(halo, storage)
    profile_group = fh.create_group("profiles")
    for field in my_profile:
        # Don't write code units because we might not know those later.
        if isinstance(my_profile[field], YTArray):
            my_profile[field].convert_to_cgs()
        _yt_array_hdf5(profile_group, str(field), my_profile[field])
    variance_storage = "%s_variance" % storage
    if hasattr(halo, variance_storage):
        my_profile = getattr(halo, variance_storage)
        variance_group = fh.create_group("variance")
        for field in my_profile:
            # Don't write code units because we might not know those later.
            if isinstance(my_profile[field], YTArray):
                my_profile[field].convert_to_cgs()
            _yt_array_hdf5(variance_group, str(field), my_profile[field])
    fh.close()

add_callback("save_profiles", save_profiles)

def load_profiles(halo, storage="profiles", fields=None,
                  filename=None, output_dir="."):
    r"""
    Load profile data from disk.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    storage : string
        Name of the dictionary attribute to store profile data.
        Default: "profiles"
    fields : string or list of strings
        The fields to be loaded.  If None, all fields present will be loaded.
        Default : None
    filename : string
        The name of the file to be loaded.  The final filename will be 
        "<filename>_<id>.h5".  If None, filename is set to the value given 
        by the storage keyword.
        Default: None
    output_dir : string
        Name of directory where profile data will be read.  The full path will be
        the output_dir of the halo catalog concatenated with this directory.
        Default : "."
    
    """

    if filename is None:
        filename = storage
    output_file = os.path.join(halo.halo_catalog.output_dir, output_dir,
                               "%s_%06d.h5" % (filename, 
                                               halo.quantities["particle_identifier"]))
    if not os.path.exists(output_file):
        raise RuntimeError("Profile file not found: %s." % output_file)
    mylog.info("Loading halo %d profile data from %s." %
               (halo.quantities["particle_identifier"], output_file))

    fh = h5py.File(output_file, "r")
    if fields is None:
        profile_fields = fh["profiles"].keys()
    else:
        profile_fields = fields
    my_profile = {}
    my_group = fh["profiles"]
    for field in profile_fields:
        if field not in my_group:
            raise RuntimeError("%s field not present in %s." % (field, output_file))
        my_profile[field] = _hdf5_yt_array(my_group, field,
                                           ds=halo.halo_catalog.halos_ds)
    setattr(halo, storage, my_profile)
    
    if "variance" in fh:
        my_variance = {}
        my_group = fh["variance"]
        if fields is None:
            profile_fields = my_group.keys()
        for field in profile_fields:
            if field not in my_group:
                raise RuntimeError("%s field not present in %s." % (field, output_file))
            my_variance[field] = _hdf5_yt_array(my_group, field,
                                                ds=halo.halo_catalog.halos_ds)
        setattr(halo, "%s_variance" % storage, my_variance)
        
    fh.close()

add_callback("load_profiles", load_profiles)

def virial_quantities(halo, fields, 
                      overdensity_field=("gas", "overdensity"),
                      critical_overdensity=200,
                      profile_storage="profiles"):
    r"""
    Calculate the value of the given fields at the virial radius defined at 
    the given critical density by interpolating from radial profiles.

    Parameters
    ----------    
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    fields : string or list of strings
        The fields whose virial values are to be calculated.
    overdensity_field : string or tuple of strings
        The field used as the overdensity from which interpolation is done to 
        calculate virial quantities.
        Default: ("gas", "overdensity")
    critical_overdensity : float
        The value of the overdensity at which to evaluate the virial quantities.  
        Overdensity is with respect to the critical density.
        Default: 200
    profile_storage : string
        Name of the halo attribute that holds the profiles to be used.
        Default: "profiles"
    
    """

    mylog.info("Calculating virial quantities for halo %d." %
               halo.quantities["particle_identifier"])

    fields = ensure_list(fields)
    fields = [halo.data_object._determine_fields(field)[0]
              for field in fields]
    
    dds = halo.halo_catalog.data_ds
    profile_data = getattr(halo, profile_storage)

    if overdensity_field not in profile_data:
      raise RuntimeError("virial_quantities callback requires profile of %s." %
                         str(overdensity_field))

    overdensity = profile_data[overdensity_field]
    dfilter = np.isfinite(overdensity) & profile_data["used"] & (overdensity > 0)

    v_fields = {}
    for field in fields:
        if isinstance(field, tuple):
            my_field = field[-1]
        else:
            my_field = field
        v_fields[field] = my_field
        v_field = "%s_%d" % (my_field, critical_overdensity)
        if v_field not in halo.halo_catalog.quantities:
            halo.halo_catalog.quantities.append(v_field)
    vquantities = dict([("%s_%d" % (v_fields[field], critical_overdensity),
                         dds.quan(0, profile_data[field].units)) \
                        for field in fields])

    if dfilter.sum() < 2:
        halo.quantities.update(vquantities)
        return

    # find interpolation index
    # require a negative slope, but not monotonicity
    vod = overdensity[dfilter].to_ndarray()
    if (vod > critical_overdensity).all():
        if vod[-1] < vod[-2]:
            index = -2
        else:
            halo.quantities.update(vquantities)
            return
    elif (vod < critical_overdensity).all():
        if vod[0] > vod[1]:
            index = 0
        else:
            halo.quantities.update(vquantities)
            return            
    else:
        # take first instance of downward intersection with critical value
        intersections = (vod[:-1] >= critical_overdensity) & \
            (vod[1:] < critical_overdensity)
        if not intersections.any():
            halo.quantities.update(vquantities)
            return            
        index = np.where(intersections)[0][0]

    for field in fields:
        v_prof = profile_data[field][dfilter].to_ndarray()
        slope = np.log(v_prof[index + 1] / v_prof[index]) / \
          np.log(vod[index + 1] / vod[index])
        value = dds.quan(np.exp(slope * np.log(critical_overdensity / 
                                               vod[index])) * v_prof[index],
                         profile_data[field].units).in_cgs()
        vquantities["%s_%d" % (v_fields[field], critical_overdensity)] = value

    halo.quantities.update(vquantities)

add_callback("virial_quantities", virial_quantities)

def phase_plot(halo, output_dir=".", phase_args=None, phase_kwargs=None):
    r"""
    Make a phase plot for the halo object.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    output_dir : string
        Name of directory where profile data will be written.  The full path will be
        the output_dir of the halo catalog concatenated with this directory.
        Default : "."
    phase_args : list
        List of arguments to be given to PhasePlot.
    phase_kwargs : dict
        Dictionary of keyword arguments to be given to PhasePlot.

    """

    if phase_args is None:
        phase_args = []
    if phase_kwargs is None:
        phase_kwargs = {}

    try:
        plot = PhasePlot(halo.data_object, *phase_args, **phase_kwargs)
        plot.save(os.path.join(halo.halo_catalog.output_dir, output_dir,
                               "halo_%06d" % halo.quantities["particle_identifier"]))
    except ValueError:
        return

add_callback("phase_plot", phase_plot)

def delete_attribute(halo, attribute):
    r"""
    Delete attribute from halo object.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    attribute : string
        The attribute to be deleted.

    """

    if hasattr(halo, attribute):
        delattr(halo, attribute)

add_callback("delete_attribute", delete_attribute)

def iterative_center_of_mass(halo, radius_field="virial_radius", inner_ratio=0.1, step_ratio=0.9,
                             units="pc"):
    r"""
    Adjust halo position by iteratively recalculating the center of mass while 
    decreasing the radius.

    Parameters
    ----------
    halo : Halo object
        The Halo object to be provided by the HaloCatalog.
    radius_field : string
        The halo quantity to be used as the radius for the sphere.
        Default: "virial_radius"
    inner_ratio : float
        The ratio of the smallest sphere radius used for calculating the center of 
        mass to the initial radius.  The sphere radius is reduced and center of mass 
        recalculated until the sphere has reached this size.
        Default: 0.1
    step_ratio : float
        The multiplicative factor used to reduce the radius of the sphere after the 
        center of mass is calculated.
        Default: 0.9
    units : str
        The units for printing out the distance between the initial and final centers.
        Default : "pc"
        
    """
    if inner_ratio <= 0.0 or inner_ratio >= 1.0:
        raise RuntimeError("iterative_center_of_mass: inner_ratio must be between 0 and 1.")
    if step_ratio <= 0.0 or step_ratio >= 1.0:
        raise RuntimeError("iterative_center_of_mass: step_ratio must be between 0 and 1.")

    center_orig = halo.halo_catalog.data_ds.arr([halo.quantities["particle_position_%s" % axis]
                                                 for axis in "xyz"])
    sphere = halo.halo_catalog.data_ds.sphere(center_orig, halo.quantities[radius_field])

    while sphere.radius > inner_ratio * halo.quantities[radius_field]:
        new_center = sphere.quantities.center_of_mass(use_gas=True, use_particles=True)
        sphere = sphere.ds.sphere(new_center, step_ratio * sphere.radius)

    distance = periodic_distance(center_orig.in_units("code_length").to_ndarray(),
                                 new_center.in_units("code_length").to_ndarray())
    distance = halo.halo_catalog.data_ds.quan(distance, "code_length")
    mylog.info("Recentering halo %d %f %s away." %
               (halo.quantities["particle_identifier"],
                distance.in_units(units), units))

    for i, axis in enumerate("xyz"):
        halo.quantities["particle_position_%s" % axis] = sphere.center[i]
    del sphere
    
add_callback("iterative_center_of_mass", iterative_center_of_mass)



import numpy as np
from nose.tools import assert_raises

from yt.testing import assert_equal, fake_amr_ds, fake_particle_ds, \
    fake_random_ds, requires_file
from yt.utilities.answer_testing.framework import data_dir_load
from yt.utilities.exceptions import YTDimensionalityError
from yt.visualization.line_plot import LineBuffer

# This will test the "dataset access" method.

def test_box_creation():
    ds = fake_random_ds(32, length_unit=2)
    left_edge = ds.arr([0.2, 0.2, 0.2], 'cm')
    right_edge = ds.arr([0.6, 0.6, 0.6], 'cm')
    center = (left_edge + right_edge)/2

    boxes = [
        ds.box(left_edge, right_edge),
        ds.box(0.5*np.array(left_edge), 0.5*np.array(right_edge)),
        ds.box((0.5*left_edge).tolist(), (0.5*right_edge).tolist())
    ]

    region = ds.region(center, left_edge, right_edge)

    for b in boxes:
        assert_equal(b.left_edge, region.left_edge)
        assert_equal(b.right_edge, region.right_edge)
        assert_equal(b.center, region.center)

def test_region_from_d():
    ds = fake_amr_ds(fields=["density"])
    # We'll do a couple here

    # First, no string units
    reg1 = ds.r[0.2:0.3,0.4:0.6,:]
    reg2 = ds.region([0.25, 0.5, 0.5], [0.2, 0.4, 0.0], [0.3, 0.6, 1.0])
    assert_equal(reg1["density"], reg2["density"])

    # Now, string units in some -- 1.0 == cm
    reg1 = ds.r[(0.1, 'cm'):(0.5, 'cm'), :, (0.25, 'cm'): (0.35, 'cm')]
    reg2 = ds.region([0.3, 0.5, 0.3], [0.1, 0.0, 0.25], [0.5, 1.0, 0.35])
    assert_equal(reg1["density"], reg2["density"])

    # Now, string units in some -- 1.0 == cm
    reg1 = ds.r[(0.1, 'cm'):(0.5, 'cm'), :, 0.25:0.35]
    reg2 = ds.region([0.3, 0.5, 0.3], [0.1, 0.0, 0.25], [0.5, 1.0, 0.35])
    assert_equal(reg1["density"], reg2["density"])

    # And, lots of : usage!
    reg1 = ds.r[:, :, :]
    reg2 = ds.all_data()
    assert_equal(reg1["density"], reg2["density"])

    # Test slice as an index
    reg1 = ds.r[0.1:0.8]
    reg2 = ds.region([0.45, 0.45, 0.45], [0.1, 0.1, 0.1], [0.8, 0.8, 0.8])
    assert_equal(reg1["density"], reg2["density"])

    # Test with bad boundary initialization
    with assert_raises(RuntimeError):
        ds.r[0.3:0.1, 0.4:0.6, :]

    # Test region by creating an arbitrary grid
    reg1 = ds.r[0.15:0.55:16j, 0.25:0.65:32j, 0.35:0.75:64j]
    left_edge = np.array([0.15, 0.25, 0.35])
    right_edge = np.array([0.55, 0.65, 0.75])
    dims = np.array([16.0, 32.0, 64.0])
    reg2 = ds.arbitrary_grid(left_edge, right_edge, dims)
    assert_equal(reg1["density"], reg2["density"])

def test_accessing_all_data():
    # This will test first that we can access all_data, and next that we can
    # access it multiple times and get the *same object*.
    ds = fake_amr_ds(fields=["density"])
    dd = ds.all_data()
    assert_equal(ds.r["density"], dd["density"])
    # Now let's assert that it's the same object
    rho = ds.r["density"]
    rho *= 2.0
    assert_equal(dd["density"]*2.0, ds.r["density"])
    assert_equal(dd["gas", "density"]*2.0, ds.r["gas", "density"])

def test_slice_from_r():
    ds = fake_amr_ds(fields = ["density"])
    sl1 = ds.r[0.5, :, :]
    sl2 = ds.slice("x", 0.5)
    assert_equal(sl1["density"], sl2["density"])

    frb1 = sl1.to_frb(width = 1.0, height = 1.0, resolution = (1024, 512))
    frb2 = ds.r[0.5, ::1024j, ::512j]
    assert_equal(frb1["density"], frb2["density"])

    # Test slice which doesn't cover the whole domain
    box = ds.box([0.0, 0.25, 0.25], [1.0, 0.75, 0.75])

    sl3 = ds.r[0.5, 0.25:0.75, 0.25:0.75]
    sl4 = ds.slice("x", 0.5, data_source=box)
    assert_equal(sl3["density"], sl4["density"])

    frb3 = sl3.to_frb(width = 0.5, height = 0.5, resolution = (1024, 512))
    frb4 = ds.r[0.5, 0.25:0.75:1024j, 0.25:0.75:512j]
    assert_equal(frb3["density"], frb4["density"])

def test_point_from_r():
    ds = fake_amr_ds(fields = ["density"])
    pt1 = ds.r[0.5,0.3,0.1]
    pt2 = ds.point([0.5,0.3,0.1])
    assert_equal(pt1["density"], pt2["density"])

    # Test YTDimensionalityError
    with assert_raises(YTDimensionalityError) as ex:
        ds.r[0.5, 0.1]
    assert_equal(str(ex.exception),
                 'Dimensionality specified was 2 but we need 3')

def test_ray_from_r():
    ds = fake_amr_ds(fields = ["density"])
    ray1 = ds.r[(0.1,0.2,0.3):(0.4,0.5,0.6)]
    ray2 = ds.ray((0.1,0.2,0.3), (0.4,0.5,0.6))
    assert_equal(ray1["density"], ray2["density"])

    ray3 = ds.r[0.5*ds.domain_left_edge:0.5*ds.domain_right_edge]
    ray4 = ds.ray(0.5*ds.domain_left_edge, 0.5*ds.domain_right_edge)
    assert_equal(ray3["density"], ray4["density"])

    start = [(0.1,"cm"), 0.2, (0.3,"cm")]
    end = [(0.5,"cm"), (0.4,"cm"), 0.6]
    ray5 = ds.r[start:end]
    start_arr = [ds.quan(0.1,"cm"), 0.2, ds.quan(0.3,"cm")]
    end_arr = [ds.quan(0.5,"cm"), ds.quan(0.4,"cm"), 0.6]
    ray6 = ds.ray(start_arr, end_arr)
    assert_equal(ray5["density"], ray6["density"])

    ray7 = ds.r[start:end:500j]
    ray8 = LineBuffer(ds, [0.1, 0.2, 0.3], [0.5, 0.4, 0.6], 500)
    assert_equal(ray7["density"], ray8["density"])

def test_ortho_ray_from_r():
    ds = fake_amr_ds(fields = ["density"])
    ray1 = ds.r[:,0.3,0.2]
    ray2 = ds.ortho_ray("x",[0.3, 0.2])
    assert_equal(ray1["density"], ray2["density"])

    # the y-coord is funny so test it too
    ray3 = ds.r[0.3,:,0.2]
    ray4 = ds.ortho_ray("y", [0.2, 0.3])
    assert_equal(ray3["density"], ray4["density"])

    # Test ray which doesn't cover the whole domain
    box = ds.box([0.25, 0.0, 0.0], [0.75, 1.0, 1.0])
    ray5 = ds.r[0.25:0.75,0.3,0.2]
    ray6 = ds.ortho_ray("x", [0.3, 0.2], data_source=box)
    assert_equal(ray5["density"], ray6["density"])

    # Test fixed-resolution rays
    ray7 = ds.r[0.25:0.75:100j,0.3,0.2]
    ray8 = LineBuffer(ds, [0.2525,0.3,0.2], [0.7475,0.3,0.2], 100)
    assert_equal(ray7["density"], ray8["density"])

def test_particle_counts():
    ds = fake_random_ds(16, particles=100)
    assert ds.particle_type_counts == {'io': 100}

    pds = fake_particle_ds(npart=128)
    assert pds.particle_type_counts == {'io': 128}


g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"
@requires_file(g30)
def test_checksum():
    assert fake_random_ds(16).checksum == 'notafile'
    assert data_dir_load(g30).checksum == '6169536e4b9f737ce3d3ad440df44c58'



"""
Definitions for RAMSES files




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

# These functions are RAMSES-specific
from yt.config import ytcfg
from yt.funcs import mylog
import re

def ramses_header(hvals):
    header = ( ('ncpu', 1, 'i'),
               ('ndim', 1, 'i'),
               ('nx', 3, 'i'),
               ('nlevelmax', 1, 'i'),
               ('ngridmax', 1, 'i'),
               ('nboundary', 1, 'i'),
               ('ngrid_current', 1, 'i'),
               ('boxlen', 1, 'd'),
               ('nout', 3, 'i')
              )
    yield header
    # TODO: REMOVE
    noutput, iout, ifout = hvals['nout']
    next_set = ( ('tout', noutput, 'd'),
                 ('aout', noutput, 'd'),
                 ('t', 1, 'd'),
                 ('dtold', hvals['nlevelmax'], 'd'),
                 ('dtnew', hvals['nlevelmax'], 'd'),
                 ('nstep',  2, 'i'),
                 ('stat', 3, 'd'),
                 ('cosm', 7, 'd'),
                 ('timing', 5, 'd'),
                 ('mass_sph', 1, 'd', True)
                 )
    yield next_set

field_aliases = {
    'standard_five':     ('Density',
                          'x-velocity',
                          'y-velocity',
                          'z-velocity',
                          'Pressure'),
    'standard_six':      ('Density',
                          'x-velocity',
                          'y-velocity',
                          'z-velocity',
                          'Pressure',
                          'Metallicity'),

}

## Regular expressions used to parse file descriptors
VERSION_RE = re.compile(r'# version: *(\d+)')
# This will match comma-separated strings, discarding whitespaces
# on the left hand side
VAR_DESC_RE = re.compile(r'\s*([^\s]+),\s*([^\s]+),\s*([^\s]+)')


## Configure family mapping
particle_families = {
    'DM': 1,
    'star': 2,
    'cloud': 3,
    'dust': 4,
    'star_tracer': -2,
    'cloud_tracer': -3,
    'dust_tracer': -4,
    'gas_tracer': 0
}

if ytcfg.has_section('ramses-families'):
    for key in particle_families.keys():
        val = ytcfg.getint('ramses-families', key, fallback=None)
        if val is not None:
            mylog.info('Changing family %s from %s to %s' % (key, particle_families[key], val))
            particle_families[key] = val



"""
These are common particle fields.




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.fields.derived_field import \
    ValidateParameter, \
    ValidateSpatial

from yt.fields.field_detector import \
    FieldDetector

from yt.funcs import \
    issue_deprecation_warning

from yt.units.yt_array import \
    uconcatenate, \
    ucross

from yt.utilities.math_utils import \
    get_sph_r_component, \
    get_sph_theta_component, \
    get_sph_phi_component, \
    get_cyl_r_component, \
    get_cyl_z_component, \
    get_cyl_theta_component, \
    get_cyl_r, get_cyl_theta, \
    get_cyl_z, \
    get_sph_theta, get_sph_phi, \
    modify_reference_frame

from .vector_operations import \
    create_magnitude_field

from .field_functions import \
    get_radius

sph_whitelist_fields = (
    'particle_velocity_x',
    'particle_velocity_y',
    'particle_velocity_z',
    'density',
    'temperature',
    'metallicity',
    'thermal_energy',
    'H_fraction',
    'He_fraction',
    'C_fraction',
    'N_fraction',
    'O_fraction',
    'Ne_fraction',
    'Mg_fraction',
    'Si_fraction',
    'Fe_fraction',
    'C_density',
    'O_density',
    'Si_density',
    'Fe_density'
)


def _field_concat(fname):
    def _AllFields(field, data):
        v = []
        for ptype in data.ds.particle_types:
            data.ds._last_freq = (ptype, None)
            if ptype == "all" or ptype in data.ds.known_filters:
                continue
            v.append(data[ptype, fname].copy())
        rv = uconcatenate(v, axis=0)
        return rv
    return _AllFields

def _field_concat_slice(fname, axi):
    def _AllFields(field, data):
        v = []
        for ptype in data.ds.particle_types:
            data.ds._last_freq = (ptype, None)
            if ptype == "all" or ptype in data.ds.known_filters:
                continue
            v.append(data[ptype, fname][:,axi])
        rv = uconcatenate(v, axis=0)
        return rv
    return _AllFields

def particle_deposition_functions(ptype, coord_name, mass_name, registry):
    unit_system = registry.ds.unit_system
    orig = set(registry.keys())
    ptype_dn = ptype.replace("_"," ").title()
    def particle_count(field, data):
        pos = data[ptype, coord_name]
        d = data.deposit(pos, method = "count")
        d = data.ds.arr(d, input_units = "cm**-3")
        return data.apply_units(d, field.units)

    registry.add_field(("deposit", "%s_count" % ptype),
             sampling_type="cell",
             function = particle_count,
             validators = [ValidateSpatial()],
             units = '',
             display_name = r"\mathrm{%s Count}" % ptype_dn)

    def particle_mass(field, data):
        pos = data[ptype, coord_name]
        pmass = data[ptype, mass_name]
        pmass.convert_to_units(field.units)
        d = data.deposit(pos, [pmass], method = "sum")
        return data.apply_units(d, field.units)

    registry.add_field(("deposit", "%s_mass" % ptype),
             sampling_type="cell",
             function = particle_mass,
             validators = [ValidateSpatial()],
             display_name = r"\mathrm{%s Mass}" % ptype_dn,
             units = unit_system["mass"])

    def particle_density(field, data):
        pos = data[ptype, coord_name].convert_to_units("code_length")
        mass = data[ptype, mass_name].convert_to_units("code_mass")
        d = data.deposit(pos, [mass], method = "sum")
        d = data.ds.arr(d, "code_mass")
        d /= data["index", "cell_volume"]
        return d

    registry.add_field(("deposit", "%s_density" % ptype),
             sampling_type="cell",
             function = particle_density,
             validators = [ValidateSpatial()],
             display_name = r"\mathrm{%s Density}" % ptype_dn,
             units = unit_system["density"])

    def particle_cic(field, data):
        pos = data[ptype, coord_name]
        d = data.deposit(pos, [data[ptype, mass_name]], method = "cic")
        d = data.apply_units(d, data[ptype, mass_name].units)
        d /= data["index", "cell_volume"]
        return d

    registry.add_field(("deposit", "%s_cic" % ptype),
             sampling_type="cell",
             function = particle_cic,
             validators = [ValidateSpatial()],
             display_name = r"\mathrm{%s CIC Density}" % ptype_dn,
             units = unit_system["density"])

    def _get_density_weighted_deposit_field(fname, units, method):
        def _deposit_field(field, data):
            """
            Create a grid field for particle quantities weighted by particle
            mass, using cloud-in-cell deposit.
            """
            pos = data[ptype, "particle_position"]
            # Get back into density
            pden = data[ptype, 'particle_mass']
            top = data.deposit(pos, [data[(ptype, fname)]*pden], method=method)
            bottom = data.deposit(pos, [pden], method=method)
            top[bottom == 0] = 0.0
            bnz = bottom.nonzero()
            top[bnz] /= bottom[bnz]
            d = data.ds.arr(top, input_units=units)
            return d
        return _deposit_field

    for ax in 'xyz':
        for method, name in zip(("cic", "sum"), ("cic", "nn")):
            function = _get_density_weighted_deposit_field(
                "particle_velocity_%s" % ax, "code_velocity", method)
            registry.add_field(
                ("deposit", ("%s_"+name+"_velocity_%s") % (ptype, ax)), sampling_type="cell",
                function=function, units=unit_system["velocity"], take_log=False,
                validators=[ValidateSpatial(0)])

    for method, name in zip(("cic", "sum"), ("cic", "nn")):
        function = _get_density_weighted_deposit_field(
            "age", "code_time", method)
        registry.add_field(
            ("deposit", ("%s_"+name+"_age") % (ptype)), sampling_type="cell",
            function=function, units=unit_system["time"], take_log=False,
            validators=[ValidateSpatial(0)])

    # Now some translation functions.

    def particle_ones(field, data):
        v = np.ones(data[ptype, coord_name].shape[0], dtype="float64")
        return data.apply_units(v, field.units)

    registry.add_field((ptype, "particle_ones"),
                       sampling_type="particle",
                       function = particle_ones,
                       units = "",
                       display_name = r"Particle Count")

    def particle_mesh_ids(field, data):
        pos = data[ptype, coord_name]
        ids = np.zeros(pos.shape[0], dtype="float64") - 1
        # This is float64 in name only.  It will be properly cast inside the
        # deposit operation.
        #_ids = ids.view("float64")
        data.deposit(pos, [ids], method = "mesh_id")
        return data.apply_units(ids, "")
    registry.add_field((ptype, "mesh_id"),
            sampling_type="particle",
            function = particle_mesh_ids,
            validators = [ValidateSpatial()],
            units = '')

    return list(set(registry.keys()).difference(orig))

def particle_scalar_functions(ptype, coord_name, vel_name, registry):

    # Now we have to set up the various velocity and coordinate things.  In the
    # future, we'll actually invert this and use the 3-component items
    # elsewhere, and stop using these.

    # Note that we pass in _ptype here so that it's defined inside the closure.

    def _get_coord_funcs(axi, _ptype):
        def _particle_velocity(field, data):
            return data[_ptype, vel_name][:,axi]
        def _particle_position(field, data):
            return data[_ptype, coord_name][:, axi]
        return _particle_velocity, _particle_position
    for axi, ax in enumerate("xyz"):
        v, p = _get_coord_funcs(axi, ptype)
        registry.add_field((ptype, "particle_velocity_%s" % ax),
            sampling_type="particle", function = v, units = "code_velocity")
        registry.add_field((ptype, "particle_position_%s" % ax),
            sampling_type="particle", function = p, units = "code_length")

def particle_vector_functions(ptype, coord_names, vel_names, registry):

    unit_system = registry.ds.unit_system

    # This will column_stack a set of scalars to create vector fields.

    def _get_vec_func(_ptype, names):
        def particle_vectors(field, data):
            v = [data[_ptype, name].in_units(field.units)
                  for name in names]
            c = np.column_stack(v)
            return data.apply_units(c, field.units)
        return particle_vectors
    registry.add_field((ptype, "particle_position"),
                       sampling_type="particle",
                       function=_get_vec_func(ptype, coord_names),
                       units = "code_length")
    registry.add_field((ptype, "particle_velocity"),
                       sampling_type="particle",
                       function=_get_vec_func(ptype, vel_names),
                       units = unit_system["velocity"])

def get_angular_momentum_components(ptype, data, spos, svel):
    if data.has_field_parameter("normal"):
        normal = data.get_field_parameter("normal")
    else:
        normal = data.ds.arr([0.0,0.0,1.0],"code_length") # default to simulation axis
    pos = data.ds.arr([data[ptype, spos % ax] for ax in "xyz"]).T
    vel = data.ds.arr(
        [data[ptype, 'relative_%s' % (svel % ax)] for ax in "xyz"]).T
    return pos, vel, normal

def standard_particle_fields(registry, ptype,
                             spos = "particle_position_%s",
                             svel = "particle_velocity_%s"):
    unit_system = registry.ds.unit_system

    def _particle_velocity_magnitude(field, data):
        """ M{|v|} """
        return np.sqrt(data[ptype, 'relative_%s' % (svel % 'x')]**2 +
                       data[ptype, 'relative_%s' % (svel % 'y')]**2 +
                       data[ptype, 'relative_%s' % (svel % 'z')]**2)

    registry.add_field((ptype, "particle_velocity_magnitude"),
                  sampling_type="particle",
                  function=_particle_velocity_magnitude,
                  take_log=False,
                  units=unit_system["velocity"])

    def _particle_specific_angular_momentum(field, data):
        """Calculate the angular of a particle velocity.

        Returns a vector for each particle.
        """
        center = data.get_field_parameter('center')
        pos, vel, normal = get_angular_momentum_components(
            ptype, data, spos, svel)
        L, r_vec, v_vec = modify_reference_frame(center, normal, P=pos, V=vel)
        # adding in the unit registry allows us to have a reference to the
        # dataset and thus we will always get the correct units after applying
        # the cross product.
        return ucross(r_vec, v_vec, registry=data.ds.unit_registry)


    registry.add_field((ptype, "particle_specific_angular_momentum"),
              sampling_type="particle",
              function=_particle_specific_angular_momentum,
              units=unit_system["specific_angular_momentum"],
              validators=[ValidateParameter("center")])

    def _get_spec_ang_mom_comp(axi, ax, _ptype):
        def _particle_specific_angular_momentum_component(field, data):
            return data[_ptype, "particle_specific_angular_momentum"][:, axi]
        def _particle_angular_momentum_component(field, data):
            return data[_ptype, "particle_mass"] * \
                data[ptype, "particle_specific_angular_momentum_%s" % ax]
        return _particle_specific_angular_momentum_component, \
            _particle_angular_momentum_component
    for axi, ax in enumerate("xyz"):
        f, v = _get_spec_ang_mom_comp(axi, ax, ptype)
        registry.add_field(
            (ptype, "particle_specific_angular_momentum_%s" % ax),
            sampling_type="particle",
            function=f,
            units=unit_system["specific_angular_momentum"],
            validators=[ValidateParameter("center")]
        )
        registry.add_field(
            (ptype, "particle_angular_momentum_%s" % ax),
            sampling_type="particle",
            function=v, units=unit_system["angular_momentum"],
            validators=[ValidateParameter('center')])

    def _particle_angular_momentum(field, data):
        am = data[ptype, "particle_mass"] * data[ptype, "particle_specific_angular_momentum"].T
        return am.T

    registry.add_field((ptype, "particle_angular_momentum"),
              sampling_type="particle",
              function=_particle_angular_momentum,
              units=unit_system["angular_momentum"],
              validators=[ValidateParameter("center")])

    create_magnitude_field(registry, "particle_angular_momentum",
                           unit_system["angular_momentum"],
                           ftype=ptype, particle_type=True)

    def _particle_radius(field, data):
        """The spherical radius component of the particle positions

        Relative to the coordinate system defined by the *normal* vector,
        and *center* field parameters.
        """
        return get_radius(data, "particle_position_", field.name[0])

    registry.add_field(
        (ptype, "particle_radius"),
        sampling_type="particle",
        function=_particle_radius,
        units=unit_system["length"],
        validators=[ValidateParameter("center")])

    def _relative_particle_position(field, data):
        """The cartesian particle positions in a rotated reference frame

        Relative to the coordinate system defined by the *normal* vector and
        *center* field parameters.

        Note that the orientation of the x and y axes are arbitrary.
        """
        normal = data.get_field_parameter('normal')
        center = data.get_field_parameter('center')
        pos = data.ds.arr([data[ptype, spos % ax] for ax in "xyz"]).T
        L, pos = modify_reference_frame(center, normal, P=pos)
        return pos

    def _particle_position_relative(field, data):
        if not isinstance(data, FieldDetector):
            issue_deprecation_warning(
                "The 'particle_position_relative' field has been deprecated in " +
                "favor of 'relative_particle_position'.")
        if isinstance(field.name, tuple):
            return data[field.name[0], 'relative_particle_position']
        else:
            return data['relative_particle_position']

    for name, func in zip(
            ["particle_position_relative", "relative_particle_position"],
            [_particle_position_relative,  _relative_particle_position]):
        registry.add_field(
            (ptype, name),
            sampling_type="particle",
            function=func,
            units=unit_system["length"],
            validators=[ValidateParameter("normal"),
                        ValidateParameter("center")])

    def _relative_particle_velocity(field, data):
        """The vector particle velocities in an arbitrary coordinate system

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.

        Note that the orientation of the x and y axes are arbitrary.
        """
        normal = data.get_field_parameter('normal')
        center = data.get_field_parameter('center')
        bv = data.get_field_parameter("bulk_velocity")
        vel = data.ds.arr(
            [data[ptype, svel % ax] - bv[iax]
             for iax, ax in enumerate("xyz")]).T
        L, vel = modify_reference_frame(center, normal, V=vel)
        return vel

    def _particle_velocity_relative(field, data):
        if not isinstance(data, FieldDetector):
            issue_deprecation_warning(
                "The 'particle_velocity_relative' field has been deprecated in " +
                "favor of 'relative_particle_velocity'.")
        if isinstance(field.name, tuple):
            return data[field.name[0], 'relative_particle_velocity']
        else:
            return data['relative_particle_velocity']

    for name, func in zip(
            ["particle_velocity_relative", "relative_particle_velocity"],
            [_particle_velocity_relative,  _relative_particle_velocity]):
        registry.add_field(
            (ptype, name),
            sampling_type="particle",
            function=func,
            units=unit_system["velocity"],
            validators=[ValidateParameter("normal"),
                        ValidateParameter("center")])


    def _get_coord_funcs_relative(axi, _ptype):
        def _particle_pos_rel(field, data):
            return data[_ptype, "relative_particle_position"][:, axi]
        def _particle_vel_rel(field, data):
            return data[_ptype, "relative_particle_velocity"][:, axi]
        return _particle_vel_rel, _particle_pos_rel
    for axi, ax in enumerate("xyz"):
        v, p = _get_coord_funcs_relative(axi, ptype)
        registry.add_field((ptype, "particle_velocity_relative_%s" % ax),
            sampling_type="particle", function = v, units = "code_velocity")
        registry.add_field((ptype, "particle_position_relative_%s" % ax),
            sampling_type="particle", function = p, units = "code_length")
        registry.add_field((ptype, "relative_particle_velocity_%s" % ax),
            sampling_type="particle", function = v, units = "code_velocity")
        registry.add_field((ptype, "relative_particle_position_%s" % ax),
            sampling_type="particle", function = p, units = "code_length")

    # this is just particle radius but we add it with an alias for the sake of
    # consistent naming
    registry.add_field((ptype, "particle_position_spherical_radius"),
              sampling_type="particle",
              function=_particle_radius,
              units=unit_system["length"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_spherical_position_radius(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, 'particle_position_spherical_radius']

    registry.add_field((ptype, "particle_spherical_position_radius"),
              sampling_type="particle",
              function=_particle_spherical_position_radius,
              units=unit_system["length"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_position_spherical_theta(field, data):
        """The spherical theta coordinate of the particle positions.

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        return data.ds.arr(get_sph_theta(pos, normal), "")

    registry.add_field(
        (ptype, "particle_position_spherical_theta"),
        sampling_type="particle",
        function=_particle_position_spherical_theta,
        units="",
        validators=[ValidateParameter("center"), ValidateParameter("normal")])

    def _particle_spherical_position_theta(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, 'particle_position_spherical_theta']

    registry.add_field((ptype, "particle_spherical_position_theta"),
              sampling_type="particle",
              function=_particle_spherical_position_theta,
              units="",
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_position_spherical_phi(field, data):
        """The spherical phi component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        return data.ds.arr(get_sph_phi(pos, normal), "")

    registry.add_field(
        (ptype, "particle_position_spherical_phi"),
        sampling_type="particle",
        function=_particle_position_spherical_phi,
        units="",
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_spherical_position_phi(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, 'particle_position_spherical_phi']

    registry.add_field((ptype, "particle_spherical_position_phi"),
             sampling_type="particle",
             function=_particle_spherical_position_phi,
             units="",
             validators=[ValidateParameter("center"),
                         ValidateParameter("normal")])

    def _particle_velocity_spherical_radius(field, data):
        """The spherical radius component of the particle velocities in an
         arbitrary coordinate system

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        vel = data['relative_particle_velocity'].T
        theta = get_sph_theta(pos, normal)
        phi = get_sph_phi(pos, normal)
        sphr = get_sph_r_component(vel, theta, phi, normal)
        return sphr

    registry.add_field((ptype, "particle_velocity_spherical_radius"),
              sampling_type="particle",
              function=_particle_velocity_spherical_radius,
              units=unit_system["velocity"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_spherical_velocity_radius(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, 'particle_velocity_spherical_radius']

    registry.add_field((ptype, "particle_spherical_velocity_radius"),
              sampling_type="particle",
              function=_particle_spherical_velocity_radius,
              units=unit_system["velocity"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    # particel_velocity_spherical_radius is simply aliased to
    # "particle_radial_velocity" for convenience
    registry.add_field((ptype, "particle_radial_velocity"),
              sampling_type="particle",
              function=_particle_spherical_velocity_radius,
              units=unit_system["velocity"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_velocity_spherical_theta(field, data):
        """The spherical theta component of the particle velocities in an
         arbitrary coordinate system

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        vel = data['relative_particle_velocity'].T
        theta = get_sph_theta(pos, normal)
        phi = get_sph_phi(pos, normal)
        spht = get_sph_theta_component(vel, theta, phi, normal)
        return spht

    registry.add_field(
        (ptype, "particle_velocity_spherical_theta"),
        sampling_type="particle",
        function=_particle_velocity_spherical_theta,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_spherical_velocity_theta(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, 'particle_velocity_spherical_theta']

    registry.add_field((ptype, "particle_spherical_velocity_theta"),
              sampling_type="particle",
              function=_particle_spherical_velocity_theta,
              units=unit_system["velocity"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_velocity_spherical_phi(field, data):
        """The spherical phi component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        vel = data['relative_particle_velocity'].T
        phi = get_sph_phi(pos, normal)
        sphp = get_sph_phi_component(vel, phi, normal)
        return sphp

    registry.add_field(
        (ptype, "particle_velocity_spherical_phi"),
        sampling_type="particle",
        function=_particle_velocity_spherical_phi,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_spherical_velocity_phi(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, 'particle_spherical_velocity_theta']

    registry.add_field((ptype, "particle_spherical_velocity_phi"),
              sampling_type="particle",
              function=_particle_spherical_velocity_phi,
              units=unit_system["velocity"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_position_cylindrical_radius(field, data):
        """The cylindrical radius component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        return data.ds.arr(get_cyl_r(pos, normal), 'code_length')

    registry.add_field(
        (ptype, "particle_position_cylindrical_radius"),
        sampling_type="particle",
        function=_particle_position_cylindrical_radius,
        units=unit_system["length"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_position_cylindrical_theta(field,data):
        """The cylindrical theta component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        return data.ds.arr(get_cyl_theta(pos, normal), "")

    registry.add_field(
        (ptype, "particle_position_cylindrical_theta"),
        sampling_type="particle",
        function=_particle_position_cylindrical_theta,
        units="",
        validators=[ValidateParameter("center"), ValidateParameter("normal")])

    def _particle_position_cylindrical_z(field,data):
        """The cylindrical z component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        return data.ds.arr(get_cyl_z(pos, normal), 'code_length')

    registry.add_field(
        (ptype, "particle_position_cylindrical_z"),
        sampling_type="particle",
        function=_particle_position_cylindrical_z,
        units=unit_system["length"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_velocity_cylindrical_radius(field, data):
        """The cylindrical radius component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        vel = data['relative_particle_velocity'].T
        theta = get_cyl_theta(pos, normal)
        cylr = get_cyl_r_component(vel, theta, normal)
        return cylr

    registry.add_field(
        (ptype, "particle_velocity_cylindrical_radius"),
        sampling_type="particle",
        function=_particle_velocity_cylindrical_radius,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_velocity_cylindrical_theta(field, data):
        """The cylindrical theta component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        pos = data['relative_particle_position'].T
        vel = data['relative_particle_velocity'].T
        theta = get_cyl_theta(pos, normal)
        cylt = get_cyl_theta_component(vel, theta, normal)
        return cylt

    registry.add_field(
        (ptype, "particle_velocity_cylindrical_theta"),
        sampling_type="particle",
        function=_particle_velocity_cylindrical_theta,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_cylindrical_velocity_theta(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, 'particle_velocity_cylindrical_theta']

    registry.add_field((ptype, "particle_cylindrical_velocity_theta"),
              sampling_type="particle",
              function=_particle_cylindrical_velocity_theta,
              units="cm/s",
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])

    def _particle_velocity_cylindrical_z(field, data):
        """The cylindrical z component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter('normal')
        vel = data['relative_particle_velocity'].T
        cylz = get_cyl_z_component(vel, normal)
        return cylz

    registry.add_field(
        (ptype, "particle_velocity_cylindrical_z"),
        sampling_type="particle",
        function=_particle_velocity_cylindrical_z,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")])

    def _particle_cylindrical_velocity_z(field, data):
        """This field is deprecated and will be removed in a future release"""
        return data[ptype, "particle_velocity_cylindrical_z"]

    registry.add_field((ptype, "particle_cylindrical_velocity_z"),
              sampling_type="particle",
              function=_particle_cylindrical_velocity_z,
              units=unit_system["velocity"],
              validators=[ValidateParameter("normal"),
                          ValidateParameter("center")])


def add_particle_average(registry, ptype, field_name,
                         weight = "particle_mass",
                         density = True):
    field_units = registry[ptype, field_name].units
    def _pfunc_avg(field, data):
        pos = data[ptype, "particle_position"]
        f = data[ptype, field_name]
        wf = data[ptype, weight]
        f *= wf
        v = data.deposit(pos, [f], method = "sum")
        w = data.deposit(pos, [wf], method = "sum")
        v /= w
        if density: v /= data["index", "cell_volume"]
        v[np.isnan(v)] = 0.0
        return v
    fn = ("deposit", "%s_avg_%s" % (ptype, field_name))
    registry.add_field(fn,
                       sampling_type="cell",
                       function=_pfunc_avg,
                       validators = [ValidateSpatial(0)],
                       units = field_units)
    return fn

def add_volume_weighted_smoothed_field(ptype, coord_name, mass_name,
        smoothing_length_name, density_name, smoothed_field, registry,
        nneighbors = 64, kernel_name = 'cubic'):
    unit_system = registry.ds.unit_system
    if kernel_name == 'cubic':
        field_name = ("deposit", "%s_smoothed_%s" % (ptype, smoothed_field))
    else:
        field_name = ("deposit", "%s_%s_smoothed_%s" % (ptype, kernel_name,
                      smoothed_field))
    field_units = registry[ptype, smoothed_field].units
    def _vol_weight(field, data):
        pos = data[ptype, coord_name]
        pos = pos.convert_to_units("code_length")
        mass = data[ptype, mass_name].in_base(unit_system.name)
        dens = data[ptype, density_name].in_base(unit_system.name)
        quan = data[ptype, smoothed_field]
        if hasattr(quan, "units"):
            quan = quan.convert_to_units(field_units)

        if smoothing_length_name is None:
            hsml = np.zeros(quan.shape, dtype='float64') - 1
            hsml = data.apply_units(hsml, "code_length")
        else:
            hsml = data[ptype, smoothing_length_name]
            hsml.convert_to_units("code_length")
        # This is for applying cutoffs, similar to in the SPLASH paper.
        smooth_cutoff = data["index","cell_volume"]**(1./3)
        smooth_cutoff.convert_to_units("code_length")
        # volume_weighted smooth operations return lists of length 1.
        rv = data.smooth(pos, [mass, hsml, dens, quan],
                         index_fields=[smooth_cutoff],
                         method="volume_weighted",
                         create_octree=True,
                         nneighbors=nneighbors,
                         kernel_name=kernel_name)[0]
        rv[np.isnan(rv)] = 0.0
        # Now some quick unit conversions.
        # This should be used when seeking a non-normalized value:
        rv /= hsml.uq**3 / hsml.uq.in_base(unit_system.name).uq**3
        rv = data.apply_units(rv, field_units)
        return rv
    registry.add_field(field_name,
                       sampling_type="cell",
                       function = _vol_weight,
                       validators = [ValidateSpatial(0)],
                       units = field_units)
    registry.find_dependencies((field_name,))
    return [field_name]

def add_nearest_neighbor_field(ptype, coord_name, registry, nneighbors = 64):
    field_name = (ptype, "nearest_neighbor_distance_%s" % (nneighbors))
    def _nth_neighbor(field, data):
        pos = data[ptype, coord_name]
        pos.convert_to_units("code_length")
        distances = 0.0 * pos[:,0]
        data.particle_operation(pos, [distances],
                         method="nth_neighbor",
                         nneighbors = nneighbors)
        # Now some quick unit conversions.
        return distances
    registry.add_field(field_name,
                       sampling_type="particle",
                       function = _nth_neighbor,
                       validators = [ValidateSpatial(0)],
                       units = "code_length")
    return [field_name]

def add_nearest_neighbor_value_field(ptype, coord_name, sampled_field, registry):
    """
    This adds a nearest-neighbor field, where values on the mesh are assigned
    based on the nearest particle value found.  This is useful, for instance,
    with voronoi-tesselations.
    """
    field_name = ("deposit", "%s_nearest_%s" % (ptype, sampled_field))
    field_units = registry[ptype, sampled_field].units
    unit_system = registry.ds.unit_system
    def _nearest_value(field, data):
        pos = data[ptype, coord_name]
        pos = pos.convert_to_units("code_length")
        value = data[ptype, sampled_field].in_base(unit_system.name)
        rv = data.smooth(pos, [value],
                         method="nearest",
                         create_octree=True,
                         nneighbors=1)
        rv = data.apply_units(rv, field_units)
        return rv
    registry.add_field(field_name, sampling_type="cell",
            function=_nearest_value, validators=[ValidateSpatial(0)],
            units=field_units)
    return [field_name]

def add_union_field(registry, ptype, field_name, units):
    """
    Create a field that is the concatenation of multiple particle types.
    This allows us to create fields for particle unions using alias names.
    """

    def _cat_field(field, data):
        return uconcatenate([data[dep_type, field_name]
                             for dep_type in data.ds.particle_types_raw])

    registry.add_field((ptype, field_name),
                       sampling_type="particle",
                       function=_cat_field,
                       units=units)



import numpy as np

from .definitions import (
    gadget_field_specs,
    gadget_ptype_specs,
)
from .data_structures import (
    GadgetBinaryHeader,
    GadgetDataset,
)
from .io import IOHandlerGadgetBinary


vector_fields = dict(IOHandlerGadgetBinary._vector_fields)

block_ids = {
    'Coordinates': 'POS',
    'Velocities': 'VEL',
    'ParticleIDs': 'ID',
    'Mass': 'MASS',
    'InternalEnergy': 'U',
    'Density': 'RHO',
    'SmoothingLength': 'HSML',
}


def write_record(fp, data, endian):
    dtype = endian + 'i4'
    size = np.array(data.nbytes, dtype=dtype)
    fp.write(size.tobytes())
    fp.write(data.tobytes())
    fp.write(size.tobytes())


def write_block(fp, data, endian, fmt, block_id):
    assert fmt in [1, 2]
    block_id = '%-4s' % block_id
    if fmt ==2:
        block_id_dtype = np.dtype(
            [('id', 'S', 4), ('offset', endian + 'i4', 1)]
        )
        block_id_data = np.zeros(1, dtype=block_id_dtype)
        block_id_data['id'] = block_id
        block_id_data['offset'] = data.nbytes + 8
        write_record(fp, block_id_data, endian)
    write_record(fp, data, endian)


def fake_gadget_binary(
        filename='fake_gadget_binary',
        npart=(100, 100, 100, 0, 100, 0),
        header_spec='default',
        field_spec='default',
        ptype_spec='default',
        endian='', fmt=2
    ):
    """Generate a fake Gadget binary snapshot."""
    header = GadgetBinaryHeader(filename, header_spec)
    field_spec = GadgetDataset._setup_binary_spec(
        field_spec, gadget_field_specs
    )
    ptype_spec = GadgetDataset._setup_binary_spec(
        ptype_spec, gadget_ptype_specs
    )
    with open(filename, 'wb') as fp:
        # Generate and write header blocks
        for i_header, header_spec in enumerate(header.spec):
            header_dtype = np.dtype(
                [(name, endian + dtype, dim) for name, dim, dtype in header_spec]
            )
            header = np.zeros(1, dtype=header_dtype)
            if i_header == 0:
                header['Npart'] = npart
                header['Nall'] = npart
                header['NumFiles'] = 1
                header['BoxSize'] = 1
                header['HubbleParam'] = 1
            write_block(fp, header, endian, fmt, 'HEAD')

        npart = dict(zip(ptype_spec, npart))
        for fs in field_spec:
            # Parse field name and particle type
            if isinstance(fs, str):
                field = fs
                ptype = ptype_spec
            else:
                field, ptype = fs
                if isinstance(ptype, str):
                    ptype = (ptype,)
            # Determine field dimension
            if field in vector_fields:
                dim = vector_fields[field]
            else:
                dim = 1
            # Determine dtype (in numpy convention)
            if field == 'ParticleIDs':
                dtype = 'u4'
            else:
                dtype = 'f4'
            dtype = endian + dtype
            # Generate and write field block
            data = []
            for pt in ptype:
                data += [np.random.rand(npart[pt], dim)]
            data = np.concatenate(data).astype(dtype)
            if field in block_ids:
                block_id = block_ids[field]
            else:
                block_id = ''
            write_block(fp, data, endian, fmt, block_id)
    return filename



"""Drop-in replacement for the thread module.

Meant to be used as a brain-dead substitute so that threaded code does
not need to be rewritten for when the thread module is not present.

Suggested usage is::

    try:
        import _thread
    except ImportError:
        import _dummy_thread as _thread

"""
# Exports only things specified by thread documentation;
# skipping obsolete synonyms allocate(), start_new(), exit_thread().
__all__ = ['error', 'start_new_thread', 'exit', 'get_ident', 'allocate_lock',
           'interrupt_main', 'LockType']

# A dummy value
TIMEOUT_MAX = 2**31

# NOTE: this module can be imported early in the extension building process,
# and so top level imports of other modules should be avoided.  Instead, all
# imports are done when needed on a function-by-function basis.  Since threads
# are disabled, the import lock should not be an issue anyway (??).

class error(Exception):
    """Dummy implementation of _thread.error."""

    def __init__(self, *args):
        self.args = args

def start_new_thread(function, args, kwargs={}):
    """Dummy implementation of _thread.start_new_thread().

    Compatibility is maintained by making sure that ``args`` is a
    tuple and ``kwargs`` is a dictionary.  If an exception is raised
    and it is SystemExit (which can be done by _thread.exit()) it is
    caught and nothing is done; all other exceptions are printed out
    by using traceback.print_exc().

    If the executed function calls interrupt_main the KeyboardInterrupt will be
    raised when the function returns.

    """
    if type(args) != type(tuple()):
        raise TypeError("2nd arg must be a tuple")
    if type(kwargs) != type(dict()):
        raise TypeError("3rd arg must be a dict")
    global _main
    _main = False
    try:
        function(*args, **kwargs)
    except SystemExit:
        pass
    except Exception:
        import traceback
        traceback.print_exc()
    _main = True
    global _interrupt
    if _interrupt:
        _interrupt = False
        raise KeyboardInterrupt

def exit():
    """Dummy implementation of _thread.exit()."""
    raise SystemExit

def get_ident():
    """Dummy implementation of _thread.get_ident().

    Since this module should only be used when _threadmodule is not
    available, it is safe to assume that the current process is the
    only thread.  Thus a constant can be safely returned.
    """
    return -1

def allocate_lock():
    """Dummy implementation of _thread.allocate_lock()."""
    return LockType()

def stack_size(size=None):
    """Dummy implementation of _thread.stack_size()."""
    if size is not None:
        raise error("setting thread stack size not supported")
    return 0

class LockType(object):
    """Class implementing dummy implementation of _thread.LockType.

    Compatibility is maintained by maintaining self.locked_status
    which is a boolean that stores the state of the lock.  Pickling of
    the lock, though, should not be done since if the _thread module is
    then used with an unpickled ``lock()`` from here problems could
    occur from this class not having atomic methods.

    """

    def __init__(self):
        self.locked_status = False

    def acquire(self, waitflag=None, timeout=-1):
        """Dummy implementation of acquire().

        For blocking calls, self.locked_status is automatically set to
        True and returned appropriately based on value of
        ``waitflag``.  If it is non-blocking, then the value is
        actually checked and not set if it is already acquired.  This
        is all done so that threading.Condition's assert statements
        aren't triggered and throw a little fit.

        """
        if waitflag is None or waitflag:
            self.locked_status = True
            return True
        else:
            if not self.locked_status:
                self.locked_status = True
                return True
            else:
                if timeout > 0:
                    import time
                    time.sleep(timeout)
                return False

    __enter__ = acquire

    def __exit__(self, typ, val, tb):
        self.release()

    def release(self):
        """Release the dummy lock."""
        # XXX Perhaps shouldn't actually bother to test?  Could lead
        #     to problems for complex, threaded code.
        if not self.locked_status:
            raise error
        self.locked_status = False
        return True

    def locked(self):
        return self.locked_status

# Used to signal that interrupt_main was called in a "thread"
_interrupt = False
# True when not executing in a "thread"
_main = True

def interrupt_main():
    """Set _interrupt flag to True to have start_new_thread raise
    KeyboardInterrupt upon exiting."""
    if _main:
        raise KeyboardInterrupt
    else:
        global _interrupt
        _interrupt = True



"""
Data structures for AdaptaHOP



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------


HEADER_ATTRIBUTES = (
    ('npart', 1, 'i'),
    ('massp', 1, 'f'),
    ('aexp', 1, 'f'),
    ('omega_t', 1, 'f'),
    ('age', 1, 'f'),
    (('nhalos', 'nsubs'), 2, 'i')
)

HALO_ATTRIBUTES = (
    ('npart', 1, 'i'),
    ('particle_identities', -1, 'i'),
    ('particle_identifier', 1, 'i'),
    ('timestep', 1 , 'i'),
    (('level', 'host_id', 'first_subhalo_id', 'n_subhalos', 'next_subhalo_id'), 5, 'i'),
    ('particle_mass', 1, 'f'),
    (('raw_position_x', 'raw_position_y', 'raw_position_z'), 3, 'f'),
    (('particle_velocity_x', 'particle_velocity_y', 'particle_velocity_z'), 3, 'f'),
    (('particle_angular_momentum_x', 'particle_angular_momentum_y', 'particle_angular_momentum_z'), 3, 'f'),
    (('r', 'a', 'b', 'c'), 4, 'f'),
    (('ek', 'ep', 'etot'), 3, 'f'),
    ('spin', 1 , 'f'),
    (('virial_radius', 'virial_mass', 'virial_temperature', 'virial_velocity'), 4, 'f'),
    (('rho0', 'R_c'), 2, 'f')
)



"""
halo_catalog frontend tests



"""

#-----------------------------------------------------------------------------
# Copyright (c) yt Development Team. All rights reserved.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np

from yt.convenience import \
    load as yt_load
from yt.frontends.halo_catalog.data_structures import \
    HaloCatalogDataset
from yt.frontends.ytdata.utilities import \
    save_as_dataset
from yt.testing import \
    assert_array_equal, \
    requires_module, \
    TempDirTest
from yt.units.yt_array import \
    YTArray, \
    YTQuantity

def fake_halo_catalog(data):
    filename = "catalog.0.h5"

    ftypes = dict((field, '.') for field in data)
    extra_attrs = {"data_type": "halo_catalog",
                   "num_halos": data['particle_mass'].size}

    ds = {'cosmological_simulation': 1,
          'omega_lambda': 0.7,
          'omega_matter': 0.3,
          'hubble_constant': 0.7,
          'current_redshift': 0,
          'current_time': YTQuantity(1, 'yr'),
          'domain_left_edge': YTArray(np.zeros(3), 'cm'),
          'domain_right_edge': YTArray(np.ones(3), 'cm')}
    save_as_dataset(
        ds, filename, data,
        field_types=ftypes, extra_attrs=extra_attrs)
    return filename

class HaloCatalogTest(TempDirTest):
    @requires_module('h5py')
    def test_halo_catalog(self):
        rs = np.random.RandomState(3670474)
        n_halos = 100
        fields = ['particle_%s' % name for name in
                  ['mass'] + ['position_%s' % ax for ax in 'xyz']]
        units = ['g'] + ['cm']*3
        data = dict((field, YTArray(rs.random_sample(n_halos), unit))
                    for field, unit in zip(fields, units))

        fn = fake_halo_catalog(data)
        ds = yt_load(fn)

        assert isinstance(ds, HaloCatalogDataset)

        for field in fields:
            f1 = data[field].in_base()
            f1.sort()
            f2 = ds.r[field].in_base()
            f2.sort()
            assert_array_equal(f1, f2)

    @requires_module('h5py')
    def test_halo_catalog_boundary_particles(self):
        rs = np.random.RandomState(3670474)
        n_halos = 100
        fields = ['particle_%s' % name for name in
                  ['mass'] + ['position_%s' % ax for ax in 'xyz']]
        units = ['g'] + ['cm']*3
        data = dict((field, YTArray(rs.random_sample(n_halos), unit))
                    for field, unit in zip(fields, units))

        data['particle_position_x'][0] = 1.0
        data['particle_position_x'][1] = 0.0
        data['particle_position_y'][2] = 1.0
        data['particle_position_y'][3] = 0.0
        data['particle_position_z'][4] = 1.0
        data['particle_position_z'][5] = 0.0

        fn = fake_halo_catalog(data)
        ds = yt_load(fn)

        assert isinstance(ds, HaloCatalogDataset)

        for field in ['particle_mass']:
            f1 = data[field].in_base()
            f1.sort()
            f2 = ds.r[field].in_base()
            f2.sort()
            assert_array_equal(f1, f2)



"""
Two Point Functions Framework.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.utilities.on_demand_imports import _h5py as h5py
import numpy as np

from yt.funcs import mylog
from yt.utilities.performance_counters import yt_counters
from yt.utilities.parallel_tools.parallel_analysis_interface import ParallelAnalysisInterface, parallel_blocking_call, parallel_root_only

try:
    from yt.utilities.kdtree.api import \
        fKD, free_tree, create_tree
except ImportError:
    mylog.debug("The Fortran kD-Tree did not import correctly.")

import math
import inspect
import time
from collections import defaultdict

sep = 12

class TwoPointFunctions(ParallelAnalysisInterface):
    r""" Initialize a two point functions object.
    
    Parameters
    ----------
    total_values : Integer
        How many total (global) pair calculations to run for each of the
        functions specified. Default: 1000000.
    comm_size : Integer
        How entries are sent during communication. Default: 10000.
    length_type : String
        Controls the even spacing of the rulers lengths in
        logarithmic or linear space, set by "log" or "lin", respectively.
        Default: "lin".
    length_number : Integer
        Sets how many lengths to create, evenly spaced by the above
        parameter. Default: 10.
    length_range : Float
        A min/max pair for the range of values to search the over
        the simulated volume. Default: [sqrt(3)dx, 1/2*shortest box edge],
        where dx is the smallest grid cell size.
    vol_ratio : Integer
        How to multiply-assign subvolumes to the parallel
        tasks. This number must be an integer factor of the total number of tasks or
        very bad things will happen. The default value of 1 will assign one task
        to each subvolume, and there will be an equal number of subvolumes as tasks.
        A value of 2 will assign two tasks to each subvolume and there will be
        one-half as many subvolumes as tasks.
        A value equal to the number of parallel tasks will result in each task
        owning a complete copy of all the fields data, meaning each task will be
        operating on the identical full volume.
        Setting it to -1 will automatically adjust it such that each task
        owns the entire volume. Default = 1.
    salt : Integer
        A number that will be added to the random number generator
        seed. Use this if a different random series of numbers is desired when
        keeping everything else constant from this set: (MPI task count, 
        number of ruler lengths, ruler min/max, number of functions,
        number of point pairs per ruler length). Default = 0.
    theta : Float
        For random pairs of points, the second point is found by traversing
        a distance along a ray set by the angle (phi, theta) from the first
        point. To keep this angle constant, set ``theta`` to a value in the
        range [0, pi]. Default = None, which will randomize theta for
        every pair of points.
    phi : Float
        Similar to theta above, but the range of values is [0, 2*pi).
        Default = None, which will randomize phi for every pair of points.
    
    Examples
    --------
    >>> tpf = TwoPointFunctions(ds, ["velocity_x", "velocity_y", "velocity_z"],
    ... total_values=1e5, comm_size=10000, 
    ... length_number=10, length_range=[1./128, .5],
    ... length_type="log")
    """
    def __init__(self, ds, fields, left_edge=None, right_edge=None,
            total_values=1000000, comm_size=10000, length_type="lin",
            length_number=10, length_range=None, vol_ratio = 1,
            salt=0, theta=None, phi=None):
        ParallelAnalysisInterface.__init__(self)
        try:
            fKD
        except NameError:
            raise ImportError("You need to install the Forthon kD-Tree")
        self._fsets = []
        self.fields = fields
        self.constant_theta = theta
        self.constant_phi = phi
        # MPI stuff.
        self.size = self.comm.size
        self.mine = self.comm.rank
        self.vol_ratio = vol_ratio
        if self.vol_ratio == -1:
            self.vol_ratio = self.size
        self.total_values = int(total_values / self.size)
        # For communication.
        self.recv_hooks = []
        self.send_hooks = []
        self.done_hooks = []
        self.comm_size = min(int(comm_size), self.total_values)
        self.ds = ds
        self.nlevels = ds.index.max_level
        self.period = self.ds.domain_right_edge - self.ds.domain_left_edge
        self.min_edge = min(self.period)
        self.index = ds.index
        self.center = (ds.domain_right_edge + ds.domain_left_edge)/2.0
        # Figure out the range of ruler lengths.
        if length_range is None:
            length_range = [math.sqrt(3) * self.ds.index.get_smallest_dx(),
                self.min_edge/2.]
        else:
            if len(length_range) != 2:
                raise ValueError("length_range must have two values.")
            if length_range[1] <= length_range[0]:
                raise ValueError("length_range[1] must be larger than length_range[0]")
            if length_range[1] > self.min_edge/2.:
                length_range[1] = self.min_edge/2.
                mylog.info("Automatically adjusting length_range[1] to half the shortest box edge.")
        if length_range[0] == -1 or length_range[0] == -1.:
            mylog.info("Automatically adjusting length_range[0] to %1.5e." % \
                (math.sqrt(3) * self.ds.index.get_smallest_dx()))
            length_range[0] = math.sqrt(3) * self.ds.index.get_smallest_dx()
        # Make the list of ruler lengths.
        if length_type == "lin":
            self.lengths = np.linspace(length_range[0], length_range[1],
                length_number)
        elif length_type == "log":
            self.lengths = np.logspace(math.log10(length_range[0]),
                math.log10(length_range[1]), length_number)
        else:
            # Something went wrong.
            raise SyntaxError("length_type is either \"lin\" or \"log\".")
        # Subdivide the volume.
        if not left_edge or not right_edge:
            self.left_edge = self.ds.domain_left_edge
            self.right_edge = self.ds.domain_right_edge
            # This ds business below has to do with changes made for halo
            # finding on subvolumes and serves no purpose here except
            # compatibility. This is not the best policy, if I'm honest.
            ds = ds.region([0.]*3, self.left_edge, self.right_edge)
            padded, self.LE, self.RE, self.ds = \
            self.partition_index_3d(ds = ds, padding=0.,
                rank_ratio = self.vol_ratio)
        else:
            self.left_edge = left_edge
            self.right_edge = right_edge
            # We do this twice, first with no 'buffer' to get the unbuffered
            # self.LE/RE, and then second to get a buffered self.ds.
            padded, self.LE, self.RE, temp = \
                self.partition_region_3d(left_edge, right_edge,
                    rank_ratio=self.vol_ratio)
            padded, temp, temp, self.ds = \
                self.partition_region_3d(left_edge - self.lengths[-1], \
                right_edge + self.lengths[-1], rank_ratio=self.vol_ratio)
        mylog.info("LE %s RE %s %s" % (str(self.LE), str(self.RE), str(self.ds)))
        self.width = self.ds.right_edge - self.ds.left_edge
        self.mt = np.random.mtrand.RandomState(seed = 1234 * self.mine + salt)
    
    def add_function(self, function, out_labels, sqrt, corr_norm=None):
        r"""Add a function to the list that will be evaluated at the
        generated pairs of points.
        
        Parameters
        ----------
        function : Function
            The two point function of the form fcn(a, b, r1, r2, vec).
        out_labels : List of strings
            A list of strings labeling the outputs of the function.
        sqrt : List of booleans
            A list of booleans which when True will square-root the corresponding
            element of the output in the text output (write_out_means()).
        corr_norm : Float
            Used when calculating two point correlations. If set, the output
            of the function is divided by this number. Default = None.
        
        Examples
        --------
        >>> f1 = tpf.add_function(function=rms_vel, out_labels=['RMSvdiff'],
        ... sqrt=[True])
        """
        fargs = inspect.getargspec(function)
        if len(fargs.args) != 5:
            raise SyntaxError("The function %s needs five arguments." %\
                function.__name__)
        out_labels = list(out_labels)
        if len(out_labels) < 1:
            raise SyntaxError("Please specify at least one out_labels for function %s." %\
                function.__name__)
        sqrt = list(sqrt)
        if len(sqrt) != len(out_labels):
            raise SyntaxError("Please have the same number of elements in out_labels as in sqrt for function %s." %\
                function.__name__)
        self._fsets.append(FcnSet(self, function, self.min_edge,
            out_labels, sqrt,corr_norm))
        return self._fsets[-1]

    def __getitem__(self, key):
        return self._fsets[key]
    
    def run_generator(self):
        r"""After all the functions have been added, run the generator.
        
        Examples
        --------
        >>> tpf.run_generator()
        """
        yt_counters("run_generator")
        # We need a function!
        if len(self._fsets) == 0:
            mylog.error("You need to add at least one function!")
            return None
        # Do all the startup tasks to get the grid points.
        if self.nlevels == 0:
            yt_counters("build_sort")
            self._build_sort_array()
            self.sort_done = False
            yt_counters("build_sort")
        else:
            yt_counters("init_kd_tree")
            self._init_kd_tree()
            self.sort_done = True
            yt_counters("init_kd_tree")
        # Store the fields.
        self.stored_fields = {}
        yt_counters("getting data")
        for field in self.fields:
            self.stored_fields[field] = self.ds[field].copy()
        self.ds.clear_data()
        # If the arrays haven't been sorted yet and need to be, do that.
        if not self.sort_done:
            for field in self.fields:
                self.stored_fields[field] = self.stored_fields[field][self.sort]
            del self.sort
            self.sort_done = True
        yt_counters("getting data")
        self._build_fields_vals()
        yt_counters("big loop over lengths")
        t_waiting = 0.
        for bigloop, length in enumerate(self.lengths):
            self._build_points_array()
            if self.mine == 0:
                mylog.info("Doing length %1.5e" % length)
            # Things stop when this value below equals total_values.
            self.generated_points = 0
            self.gen_array = np.zeros(self.size, dtype='int64')
            self.comm_cycle_count = 0
            self.final_comm_cycle_count = 0
            self.sent_done = False
            self._setup_done_hooks_on_root()
            # While everyone else isn't done or I'm not done, we loop.
            while self._should_cycle():
                self._setup_recv_arrays()
                self._send_arrays()
                t0 = time.time()
                self.comm.mpi_Request_Waitall(self.send_hooks)
                self.comm.mpi_Request_Waitall(self.recv_hooks)
                t1 = time.time()
                t_waiting += (t1-t0)
                if (self.recv_points < -1.).any() or (self.recv_points > 1.).any(): # or \
                    #(np.abs(np.log10(np.abs(self.recv_points))) > 20).any():
                    raise ValueError("self.recv_points is no good!")
                self.points = self.recv_points.copy()
                self.fields_vals = self.recv_fields_vals.copy()
                self.gen_array = self.recv_gen_array.copy()
                self._eval_points(length)
                self.gen_array[self.mine] = self.generated_points
                self.comm_cycle_count += 1
                if self.generated_points == self.total_values:
                    self._send_done_to_root()
            if self.mine == 0:
                mylog.info("Length (%d of %d) %1.5e took %d communication cycles to complete." % \
                (bigloop+1, len(self.lengths), length, self.comm_cycle_count))
        yt_counters("big loop over lengths")
        if self.nlevels >= 1:
            del fKD.pos, fKD.qv_many, fKD.nn_tags
            free_tree(0) # Frees the kdtree object.
        yt_counters("allsum")
        self._allsum_bin_hits()
        mylog.info("Spent %f seconds waiting for communication." % t_waiting)
        yt_counters("allsum")
        yt_counters("run_generator")
    
    def _init_kd_tree(self):
        """
        Builds the kd tree of grid center points.
        """
        # Grid cell centers.
        mylog.info("Multigrid: Building kD-Tree.")
        xp = self.ds["x"]
        yp = self.ds["y"]
        zp = self.ds["z"]
        fKD.pos = np.asfortranarray(np.empty((3,xp.size), dtype='float64'))
        # Normalize the grid points only within the kdtree.
        fKD.pos[0, :] = xp[:] / self.period[0]
        fKD.pos[1, :] = yp[:] / self.period[1]
        fKD.pos[2, :] = zp[:] / self.period[2]
        fKD.nn = 1
        fKD.sort = False
        fKD.rearrange = True
        create_tree(0)

    def _build_sort_array(self):
        """
        When running on a unigrid simulation, the kD tree isn't necessary.
        But we need to ensure that the points are sorted in the usual manner
        allowing values to be found via array indices.
        """
        mylog.info("Unigrid: finding cell centers.")
        xp = self.ds["x"]
        yp = self.ds["y"]
        zp = self.ds["z"]
        self.sizes = [np.unique(xp).size, np.unique(yp).size, np.unique(zp).size]        
        self.sort = np.lexsort([zp, yp, xp])
        del xp, yp, zp
        self.ds.clear_data()
    
    def _build_fields_vals(self):
        """
        Builds an array to store the field values array.
        """
        self.fields_vals = np.empty((self.comm_size, len(self.fields)*2), \
            dtype='float64')
        # At the same time build a dict to label the columns.
        self.fields_columns = {}
        for i,field in enumerate(self.fields):
            self.fields_columns[field] = i

    def _build_points_array(self):
        """
        Initializes the array that contains the random points as all negatives
        to start with.
        """
        self.points = np.ones((self.comm_size, 6), dtype='float64') * -1.0
    
    def _setup_done_hooks_on_root(self):
        """
        Opens non-blocking receives on root pointing to all the other tasks
        """
        if self.mine != 0:
            return
        self.recv_done = {}
        for task in range(self.size):
            if task == self.mine: continue
            self.recv_done[task] = np.zeros(1, dtype='int64')
            self.done_hooks.append(self.comm.mpi_nonblocking_recv(self.recv_done[task], \
                task, tag=15))
    
    def _send_done_to_root(self):
        """
        Tell the root process that I'm done.
        """
        # If I've already done this, don't do it again.
        if self.sent_done: return
        if self.mine !=0:
            # I send when I *think* things should finish.
            self.send_done = np.ones(1, dtype='int64') * \
                (self.size / self.vol_ratio -1) + self.comm_cycle_count
            self.done_hooks.append(self.comm.mpi_nonblocking_send(self.send_done, \
                    0, tag=15))
        else:
            # As root, I need to mark myself!
            self.recv_done[0] = np.ones(1, dtype='int64') * \
                (self.size / self.vol_ratio -1) + self.comm_cycle_count
        self.sent_done = True
    
    def _should_cycle(self):
        """
        Determine if I should continue cycling the communication.
        """
        if self.mine == 0:
            # If other tasks aren't finished, this will return False.
            status = self.comm.mpi_Request_Testall(self.done_hooks)
            # Convolve this with with root's status.
            status = status * (self.generated_points == self.total_values)
            if status == 1:
                # If they are all finished, meaning Testall returns True,
                # and root has made its points, we find
                # the biggest value in self.recv_done and stop there.
                status = max(self.recv_done.values())
        else:
            status = 0
        # Broadcast the status from root - we stop only if root thinks we should
        # stop.
        status = self.comm.mpi_bcast(status)
        if status == 0: return True
        if self.comm_cycle_count < status:
            return True
        # If we've come this far, we're done.
        return False

    def _setup_recv_arrays(self):
        """
        Creates the recv buffers and calls a non-blocking MPI receive pointing
        to the left-hand neighbor.
        """
        self.recv_points = np.ones((self.comm_size, 6), dtype='float64') * -1.
        self.recv_fields_vals = np.zeros((self.comm_size, len(self.fields)*2), \
            dtype='float64')
        self.recv_gen_array = np.zeros(self.size, dtype='int64')
        self.recv_hooks.append(self.comm.mpi_nonblocking_recv(self.recv_points, \
            (self.mine-1)%self.size, tag=10))
        self.recv_hooks.append(self.comm.mpi_nonblocking_recv(self.recv_fields_vals, \
            (self.mine-1)%self.size, tag=20))
        self.recv_hooks.append(self.comm.mpi_nonblocking_recv(self.recv_gen_array, \
            (self.mine-1)%self.size, tag=40))

    def _send_arrays(self):
        """
        Send the data arrays to the right-hand neighbor.
        """
        self.send_hooks.append(self.comm.mpi_nonblocking_send(self.points,\
            (self.mine+1)%self.size, tag=10))
        self.send_hooks.append(self.comm.mpi_nonblocking_send(self.fields_vals,\
            (self.mine+1)%self.size, tag=20))
        self.send_hooks.append(self.comm.mpi_nonblocking_send(self.gen_array, \
            (self.mine+1)%self.size, tag=40))

    def _allsum_bin_hits(self):
        """
        Add up the hits to all the bins globally for all functions.
        """
        for fset in self._fsets:
            fset.too_low = self.comm.mpi_allreduce(fset.too_low, op='sum')
            fset.too_high = self.comm.mpi_allreduce(fset.too_high, op='sum')
            fset.binned = {}
            if self.mine == 0:
                mylog.info("Function %s had values out of range for these fields:" % \
                    fset.function.__name__)
                for i,field in enumerate(fset.out_labels):
                    mylog.info("Field %s had %d values too high and %d too low that were not binned." % \
                    (field, fset.too_high[i], fset.too_low[i]))
            for length in self.lengths:
                fset.length_bin_hits[length] = \
                    self.comm.mpi_allreduce(fset.length_bin_hits[length], op='sum')
                # Find out how many were successfully binned.
                fset.binned[length] = fset.length_bin_hits[length].sum()
                # Normalize the counts.
                fset.length_bin_hits[length] = \
                    fset.length_bin_hits[length].astype('float64') / \
                    fset.binned[length]
                # Return it to its original shape.
                fset.length_bin_hits[length] = \
                    fset.length_bin_hits[length].reshape(fset.bin_number)
    
    def _pick_random_points(self, length, size):
        """
        Picks out size random pairs separated by length *length*.
        """
        # First make random points inside this subvolume.
        r1 = np.empty((size,3), dtype='float64')
        for dim in range(3):
            r1[:,dim] = self.mt.uniform(low=self.ds.left_edge[dim],
                high=self.ds.right_edge[dim], size=size)
        # Next we find the second point, determined by a random
        # theta, phi angle. See Eqns. 1 & 2 from 
        # http://mathworld.wolfram.com/SpherePointPicking.html,
        # but phi and theta are switched to the Physics convention.
        if self.constant_phi is None:
            phi = self.mt.uniform(low=0, high=2.*math.pi, size=size)
        else: phi = self.constant_phi * np.ones(size, dtype='float64')
        if self.constant_theta is None:
            v = self.mt.uniform(low=0., high=1, size=size)
            theta = np.arccos(2 * v - 1)
        else: theta = self.constant_theta * np.ones(size, dtype='float64')
        r2 = np.empty((size,3), dtype='float64')
        r2[:,0] = r1[:,0] + length * np.cos(phi) * np.sin(theta)
        r2[:,1] = r1[:,1] + length * np.sin(phi) * np.sin(theta)
        r2[:,2] = r1[:,2] + length * np.cos(theta)
        # Reflect so it's inside the (full) volume.
        r2 %= self.period
        return (r1, r2)

    def _find_nearest_cell(self, points):
        """
        Finds the closest grid cell for each point in a vectorized manner.
        """
        if self.nlevels == 0:
            pos = (points - self.ds.left_edge) / self.width
            n = (self.sizes[2] * pos[:,2]).astype('int32')
            n += self.sizes[2] * (self.sizes[1] * pos[:,1]).astype('int32')
            n += self.sizes[2] * self.sizes[1] * (self.sizes[0] * pos[:,0]).astype('int32')
        else:
            # Normalize the points to a 1-period for use only within the kdtree.
            points[:, 0] = points[:, 0] / self.period[0]
            points[:, 1] = points[:, 1] / self.period[1]
            points[:, 2] = points[:, 2] / self.period[2]
            fKD.qv_many = points.T
            fKD.nn_tags = np.asfortranarray(np.empty((1, points.shape[0]), dtype='int64'))
            fKD.find_many_nn_nearest_neighbors()
            # The -1 is for fortran counting.
            n = fKD.nn_tags[0,:] - 1
        return n

    def _get_fields_vals(self, points):
        """
        Given points, return the values for the fields we need for those
        points.
        """
        # First find the grid data index field.
        indices = self._find_nearest_cell(points)
        results = np.empty((len(indices), len(self.fields)), dtype='float64')
        # Put the field values into the columns of results.
        for field in self.fields:
            col = self.fields_columns[field]
            results[:, col] = self.stored_fields[field][indices]
        return results


    def _eval_points(self, length):
        # We need to loop over the points array at least once. Further
        # iterations only happen if we have added new points to the array,
        # but not as many as we want to, so we need to check again to see if
        # we can put more points into the buffer.
        added_points = True
        while added_points:
            # If we need to, add more points to the points array.
            if self.generated_points < self.total_values:
                # Look for 'empty' slots to put in new pairs.
                select = (self.points[:,0] < 0)
                ssum = select.sum()
                # We'll generate only as many points as we need to/can.
                size = min(ssum, self.total_values - self.generated_points)
                (new_r1,new_r2) = self._pick_random_points(length, size)
                self.generated_points += size
                # If size != select.sum(), we need to pad the end of new_r1/r2
                # which is what is effectively happening below.
                newpoints = np.ones((ssum, 6), dtype='float64') * -1.
                newpoints[:size,:3] = new_r1
                newpoints[:size,3:] = new_r2
                # Now we insert them into self.points.
                self.points[select] = newpoints
            else:
                added_points = False
            
            # If we have an empty buffer here, we can skip everything below.
            if (self.points < 0).all():
                added_points = False # Not strictly required, but clearer.
                break
            
            # Now we have a points array that is either full of unevaluated points,
            # or I don't need to make any new points and I'm just processing the
            # array. Start by finding the indices of the points I own.
            self.points.shape = (self.comm_size*2, 3) # Doesn't make a copy - fast!
            select = np.bitwise_or((self.points < self.ds.left_edge).any(axis=1),
                (self.points >= self.ds.right_edge).any(axis=1))
            select = np.invert(select)
            mypoints = self.points[select]
            if mypoints.size > 0:
                # Get the fields values.
                results = self._get_fields_vals(mypoints)
                # Put this into self.fields_vals.
                self.fields_vals.shape = (self.comm_size*2, len(self.fields))
                self.fields_vals[select] = results
            
            # Put our arrays back into their original shapes cheaply!
            if mypoints.size > 0:
                self.fields_vals.shape = (self.comm_size, len(self.fields)*2)
            self.points.shape = (self.comm_size, 6)
            
            # To run the functions, what is key is that the
            # second point in the pair is ours.
            second_points = self.points[:,3:]
            select = np.bitwise_or((second_points < self.ds.left_edge).any(axis=1),
                (second_points >= self.ds.right_edge).any(axis=1))
            select = np.invert(select)
            if select.any():
                points_to_eval = self.points[select]
                fields_to_eval = self.fields_vals[select]
                
                # Find the normal vector between our points.
                vec = np.abs(points_to_eval[:,:3] - points_to_eval[:,3:])
                norm = np.sqrt(np.sum(np.multiply(vec,vec), axis=1))
                # I wish there was a better way to do this, but I can't find it.
                for i, n in enumerate(norm):
                    vec[i] = np.divide(vec[i], n)
                
                # Now evaluate the functions.
                for fcn_set in self._fsets:
                    fcn_results = fcn_set._eval_st_fcn(fields_to_eval,points_to_eval,
                        vec)
                    fcn_set._bin_results(length, fcn_results)
                
                # Now clear the buffers at the processed points.
                self.points[select] = np.array([-1.]*6, dtype='float64')
                
            else:
                # We didn't clear any points, so we should move on with our
                # lives and pass this buffer along.
                added_points = False

    @parallel_blocking_call
    def write_out_means(self, fn = "%s.txt"):
        r"""Writes out the weighted-average value for each function for
        each dimension for each ruler length to a text file. The data is written
        to files of the name 'function_name.txt' in the current working
        directory.
        
        Examples
        --------
        >>> tpf.write_out_means()
        """
        for fset in self._fsets:
            fp = self.comm.write_on_root(fn % fset.function.__name__)
            fset._avg_bin_hits()
            line = "# length".ljust(sep)
            line += "count".ljust(sep)
            for dim in fset.dims:
                line += ("%s" % fset.out_labels[dim]).ljust(sep)
            fp.write(line + "\n")
            for length in self.lengths:
                line = ("%1.5e" % length).ljust(sep)
                line += ("%d" % fset.binned[length]).ljust(sep)
                for dim in fset.dims:
                    if fset.sqrt[dim]:
                        line += ("%1.5e" % \
                            math.sqrt(fset.length_avgs[length][dim])).ljust(sep)
                    else:
                        line += ("%1.5e" % \
                            fset.length_avgs[length][dim]).ljust(sep)
                line += "\n"
                fp.write(line)
            fp.close()
    
    @parallel_root_only
    def write_out_arrays(self, fn = "%s.h5"):
        r"""Writes out the raw probability bins and the bin edges to an HDF5 file
        for each of the functions. The files are named 
        'function_name.txt' and saved in the current working directory.
        
        Examples
        --------
        >>> tpf.write_out_arrays()
        """
        if self.mine == 0:
            for fset in self._fsets:
                f = h5py.File(fn % fset.function.__name__, "w")
                bin_names = []
                prob_names = []
                bin_counts = []
                for dim in fset.dims:
                    f.create_dataset("/bin_edges_%02d_%s" % \
                        (dim, fset.out_labels[dim]), \
                        data=fset.bin_edges[dim])
                    bin_names.append("/bin_edges_%02d_%s" % \
                    (dim, fset.out_labels[dim]))
                for i,length in enumerate(self.lengths):
                    f.create_dataset("/prob_bins_%05d" % i, \
                        data=fset.length_bin_hits[length])
                    prob_names.append("/prob_bins_%05d" % i)
                    bin_counts.append([fset.too_low.sum(), fset.binned[length],
                        fset.too_high.sum()])
                f.create_dataset("/bin_edges_names", data=bin_names)
                #f.create_dataset("/prob_names", data=prob_names)
                f.create_dataset("/lengths", data=self.lengths)
                f.create_dataset("/counts", data=bin_counts)
                f.close()

    @parallel_root_only
    def write_out_correlation(self):
        r"""A special output function for doing two point correlation functions.
        Outputs the correlation function xi(r) in a text file
        'function_name_corr.txt' in the current working directory.
        
        Examples
        --------
        >>> tpf.write_out_correlation()
        """
        for fset in self._fsets:
            # Only operate on correlation functions.
            if fset.corr_norm is None: continue
            fp = self.comm.write_on_root("%s_correlation.txt" % fset.function.__name__)
            line = "# length".ljust(sep)
            line += "\\xi".ljust(sep)
            fp.write(line + "\n")
            xi = fset._corr_sum_norm()
            for length in self.lengths:
                line = ("%1.5e" % length).ljust(sep)
                line += ("%1.5e" % xi[length]).ljust(sep)
                fp.write(line + "\n")
            fp.close()

class FcnSet(TwoPointFunctions):
    def __init__(self,tpf, function, min_edge, out_labels, sqrt, corr_norm):
        self.tpf = tpf # The overarching TPF class
        self.function = function # Function to eval between the two points.
        self.min_edge = min_edge # The length of the minimum edge of the box.
        self.out_labels = out_labels # For output.
        self.sqrt = sqrt # which columns to sqrt on output.
        self.corr_norm = corr_norm # A number used to normalize a correlation function.
        # These below are used to track how many times the function returns
        # unbinned results.
        self.too_low = np.zeros(len(self.out_labels), dtype='int32')
        self.too_high = np.zeros(len(self.out_labels), dtype='int32')
        
    def set_pdf_params(self, bin_type="lin", bin_number=1000, bin_range=None):
        r"""Set the parameters used to build the Probability Distribution Function
        for each ruler length for this function. The values output by the
        function are slotted into the bins described here.
        
        Parameters
        ----------
        bin_type : String
            Controls the edges of the bins spaced evenly in
            logarithmic or linear space, set by "log" or "lin", respectively.
            A single string, or list of strings for N-dim binning.
            Default = "lin".
        bin_number : Integer
            Sets how many bins to create, evenly spaced by the above
            parameter. A single integer, or a list of integers for N-dim
            binning. Default = 1000.
        bin_range : Float
            A pair of values giving the range for the bins.
            A pair of floats (a list), or a list of pairs for N-dim binning.
            Default = None.

        Examples
        --------
        >>> f1.set_pdf_params(bin_type='log', bin_range=[5e4, 5.5e13],
        ... bin_number=1000)
        """
        # This should be called after setSearchParams.
        if not hasattr(self.tpf, "lengths"):
            mylog.error("Please call setSearchParams() before calling setPDFParams().")
            return None
        # Make sure they're either all lists or only one is.
        input = [bin_type, bin_number, bin_range]
        lists = 0
        for thing in input:
            if type(thing) == list:
                lists += 1
        if lists > 1 and lists < 3:
            mylog.error("Either all the inputs need to be lists, or only one.")
            return None
        # Make sure they're all the same length if they're lists.
        if lists == 3:
            first_len = 0
            for thing in input:
                if first_len == 0:
                    first_len = len(thing)
                    if first_len == 0:
                        mylog.error("Input cannot be an empty list.")
                        return None
                    continue
                if first_len != len(thing):
                    mylog.error("All the inputs need to have the same length.")
                    return None
        # If they are not all lists, put the input into lists for convenience.
        if lists == 1:
            bin_type, bin_number = [bin_type], [bin_number]
            bin_range = [bin_range]
        self.bin_type = bin_type
        self.bin_number = np.array(bin_number) - 1
        self.dims = range(len(bin_type))
        # Create the dict that stores the arrays to store the bin hits, and
        # the arrays themselves.
        self.length_bin_hits = {}
        for length in self.tpf.lengths:
            # It's easier to index flattened, but will be unflattened later.
            self.length_bin_hits[length] = np.zeros(self.bin_number,
                dtype='int64').flatten()
        # Create the bin edges for each dimension.
        # self.bins is indexed by dimension
        self.bin_edges = {}
        for dim in self.dims:
            # Error check.
            if len(bin_range[dim]) != 2:
                raise ValueError("bin_range must have two values.")
            if bin_range[dim][1] <= bin_range[dim][0]:
                raise ValueError("bin_range[1] must be larger than bin_range[0]")
            # Make the edges for this dimension.
            if bin_type[dim] == "lin":
                self.bin_edges[dim] = np.linspace(bin_range[dim][0], bin_range[dim][1],
                    bin_number[dim])
            elif bin_type[dim] == "log":
                self.bin_edges[dim] = np.logspace(math.log10(bin_range[dim][0]),
                    math.log10(bin_range[dim][1]), bin_number[dim])
            else:
                raise SyntaxError("bin_edges is either \"lin\" or \"log\".")

    def _eval_st_fcn(self, results, points, vec):
        """
        Return the value of the function using the provided results.
        """
        return self.function(results[:,:len(self.tpf.fields)],
            results[:,len(self.tpf.fields):], points[:,:3], points[:,3:], vec)
        """
        NOTE - A function looks like:
        def stuff(a,b,r1,r2, vec):
            return [(a[0] - b[0])/(a[1] + b[1])]
        where a and b refer to different points in space and the indices
        are for the different fields, which are given when the function is
        added. The results need to be a list or array even if it's only one
        item.
        """

    def _bin_results(self, length, results):
        """
        Add hits to the bins corresponding to these results. length_hit_bins
        is flattened, so we need to figure out the offset for this hit by
        factoring the sizes of the other dimensions.
        """
        hit_bin = np.zeros(results.shape[0], dtype='int64')
        multi = 1
        good = np.ones(results.shape[0], dtype='bool')
        for dim in range(len(self.out_labels)):
            for d1 in range(dim):
                multi *= self.bin_edges[d1].size
            if dim == 0 and len(self.out_labels)==1:
                try:
                    digi = np.digitize(results, self.bin_edges[dim])
                except ValueError:
                    # The user probably did something like 
                    # return a * b rather than
                    # return a[0] * b[0], which will only happen
                    # for single field functions.
                    digi = np.digitize(results[0], self.bin_edges[dim])
            else:
                digi = np.digitize(results[:,dim], self.bin_edges[dim])
            too_low = (digi == 0)
            too_high = (digi == self.bin_edges[dim].size)
            self.too_low[dim] += (too_low).sum()
            self.too_high[dim] += (too_high).sum()
            newgood = np.bitwise_and(np.invert(too_low), np.invert(too_high))
            good = np.bitwise_and(good, newgood)
            hit_bin += np.multiply((digi - 1), multi)
        digi_bins = np.arange(self.length_bin_hits[length].size+1)
        hist, digi_bins = np.histogram(hit_bin[good], digi_bins)
        self.length_bin_hits[length] += hist

    def _dim_sum(self, a, dim):
        """
        Given a multidimensional array a, this finds the sum over all the
        elements leaving the dimension dim untouched.
        """
        dims = np.arange(len(a.shape))
        dims = np.flipud(dims)
        gt_dims = dims[dims > dim]
        lt_dims = dims[dims < dim]
        iter_dims = np.concatenate((gt_dims, lt_dims))
        for this_dim in iter_dims:
            a = a.sum(axis=this_dim)
        return a

    def _avg_bin_hits(self):
        """
        For each dimension and length of bin_hits return the weighted average.
        """
        self.length_avgs = defaultdict(dict)
        for length in self.tpf.lengths:
            for dim in self.dims:
                self.length_avgs[length][dim] = \
                    (self._dim_sum(self.length_bin_hits[length], dim) * \
                    ((self.bin_edges[dim][:-1] + self.bin_edges[dim][1:]) / 2.)).sum()

    def _corr_sum_norm(self):
        """
        Return the correlations xi for this function. We are tacitly assuming
        that all correlation functions are one dimensional.
        """
        xi = {}
        for length in self.tpf.lengths:
            xi[length] = -1 + np.sum(self.length_bin_hits[length] * \
                self.bin_edges[0][:-1]) / self.corr_norm
        return xi



from __future__ import print_function
from __future__ import absolute_import
import numpy as np
from yt.utilities.grid_data_format.conversion.conversion_abc import Converter
from yt.utilities.on_demand_imports import _h5py as h5
from glob import glob

translation_dict = {}
translation_dict['density'] = 'density'
translation_dict['total_energy'] = 'specific_energy'
translation_dict['velocity_x'] = 'velocity_x'
translation_dict['velocity_y'] = 'velocity_y'
translation_dict['velocity_z'] = 'velocity_z'
translation_dict['cell_centered_B_x'] = 'mag_field_x'
translation_dict['cell_centered_B_y'] = 'mag_field_y'
translation_dict['cell_centered_B_z'] = 'mag_field_z'

class AthenaDistributedConverter(Converter):
    def __init__(self, basename, outname=None, source_dir=None, field_conversions=None):
        self.fields = []
        self.current_time=0.0
        name = basename.split('.')
        self.ddn = int(name[1])
        if source_dir is None:
            source_dir = './'
        self.source_dir = source_dir+'/'
        self.basename = name[0]
        if outname is None:
            outname = self.basename+'.%04i'%self.ddn+'.gdf'
        self.outname = outname
        if field_conversions is None:
            field_conversions = {}
        self.field_conversions = field_conversions
        self.handle = None

    def parse_line(self,line, grid):
        # grid is a dictionary
        splitup = line.strip().split()
        if "vtk" in splitup:
            grid['vtk_version'] = splitup[-1]
        elif "Really" in splitup:
            grid['time'] = splitup[-1]
            self.current_time = grid['time']
        elif 'PRIMITIVE' in splitup:
            grid['time'] = float(splitup[4].rstrip(','))
            grid['level'] = int(splitup[6].rstrip(','))
            grid['domain'] = int(splitup[8].rstrip(','))
            self.current_time = grid['time']
        elif "DIMENSIONS" in splitup:
            grid['dimensions'] = np.array(splitup[-3:]).astype('int')
        elif "ORIGIN" in splitup:
            grid['left_edge'] = np.array(splitup[-3:]).astype('float64')
        elif "SPACING" in splitup:
            grid['dds'] = np.array(splitup[-3:]).astype('float64')
        elif "CELL_DATA" in splitup:
            grid["ncells"] = int(splitup[-1])
        elif "SCALARS" in splitup:
            field = splitup[1]
            grid['read_field'] = field
            grid['read_type'] = 'scalar'
        elif "VECTORS" in splitup:
            field = splitup[1]
            grid['read_field'] = field
            grid['read_type'] = 'vector'

    def write_gdf_field(self, fn, grid_number, field, data):
        f = self.handle
        ## --------- Store Grid Data --------- ##
        if 'grid_%010i'%grid_number not in f['data'].keys():
            g = f['data'].create_group('grid_%010i'%grid_number)
        else:
            g = f['data']['grid_%010i'%grid_number]
        name = field
        try:
            name = translation_dict[name]
        except Exception: pass
        # print 'Writing %s' % name
        if name not in g.keys():
            g.create_dataset(name,data=data)
        


    def read_and_write_index(self,basename, ddn, gdf_name):
        """ Read Athena legacy vtk file from multiple cpus """
        proc_names = glob(self.source_dir+'id*')
        #print 'Reading a dataset from %i Processor Files' % len(proc_names)
        N = len(proc_names)
        grid_dims = np.empty([N,3],dtype='int64')
        grid_left_edges = np.empty([N,3],dtype='float64')
        grid_dds = np.empty([N,3],dtype='float64')
        grid_levels = np.zeros(N,dtype='int64')
        grid_parent_ids = -1*np.ones(N,dtype='int64')
        grid_particle_counts = np.zeros([N,1],dtype='int64')

        for i in range(N):
            if i == 0:
                fn = self.source_dir+'id%i/'%i + basename + '.%04i'%ddn + '.vtk'
            else:
                fn = self.source_dir+'id%i/'%i + basename + '-id%i'%i + '.%04i'%ddn + '.vtk'

            print('Reading file %s' % fn)
            f = open(fn,'rb')
            grid = {}
            grid['read_field'] = None
            grid['read_type'] = None
            line = f.readline()
            while grid['read_field'] is None:
                self.parse_line(line, grid)
                if "SCALAR" in line.strip().split():
                    break
                if "VECTOR" in line.strip().split():
                    break
                if 'TABLE' in line.strip().split():
                    break
                if len(line) == 0: break
                del line
                line = f.readline()

            if len(line) == 0: break
            
            if np.prod(grid['dimensions']) != grid['ncells']:
                grid['dimensions'] -= 1
                grid['dimensions'][grid['dimensions']==0]=1
            if np.prod(grid['dimensions']) != grid['ncells']:
                print('product of dimensions %i not equal to number of cells %i' % \
                      (np.prod(grid['dimensions']), grid['ncells']))
                raise TypeError

            # Append all hierarchy info before reading this grid's data
            grid_dims[i]=grid['dimensions']
            grid_left_edges[i]=grid['left_edge']
            grid_dds[i]=grid['dds']
            #grid_ncells[i]=grid['ncells']
            del grid

            f.close()
            del f
        f = self.handle 

        ## --------- Begin level nodes --------- ##
        g = f.create_group('gridded_data_format')
        g.attrs['format_version']=np.float32(1.0)
        g.attrs['data_software']='athena'
        f.create_group('data')
        f.create_group('field_types')
        f.create_group('particle_types')
        pars_g = f.create_group('simulation_parameters')


        gles = grid_left_edges
        gdims = grid_dims
        dle = np.min(gles,axis=0)
        dre = np.max(gles+grid_dims*grid_dds,axis=0)
        glis = ((gles - dle)/grid_dds).astype('int64')

        ddims = (dre-dle)/grid_dds[0]

        # grid_left_index
        f.create_dataset('grid_left_index',data=glis)
        # grid_dimensions
        f.create_dataset('grid_dimensions',data=gdims)

        # grid_level
        f.create_dataset('grid_level',data=grid_levels)

        ## ----------QUESTIONABLE NEXT LINE--------- ##
        # This data needs two dimensions for now. 
        f.create_dataset('grid_particle_count',data=grid_particle_counts)

        # grid_parent_id
        f.create_dataset('grid_parent_id',data=grid_parent_ids)

        ## --------- Done with top level nodes --------- ##

        pars_g.attrs['refine_by'] = np.int64(1)
        pars_g.attrs['dimensionality'] = np.int64(3)
        pars_g.attrs['domain_dimensions'] = ddims
        pars_g.attrs['current_time'] = self.current_time
        pars_g.attrs['domain_left_edge'] = dle
        pars_g.attrs['domain_right_edge'] = dre
        pars_g.attrs['unique_identifier'] = 'athenatest'
        pars_g.attrs['cosmological_simulation'] = np.int64(0)
        pars_g.attrs['num_ghost_zones'] = np.int64(0)
        pars_g.attrs['field_ordering'] = np.int64(1)
        pars_g.attrs['boundary_conditions'] = np.int64([0]*6) # For Now

        # Extra pars:
        # pars_g.attrs['n_cells'] = grid['ncells']
        pars_g.attrs['vtk_version'] = 1.0

        # Add particle types
        # Nothing to do here

        # Add particle field attributes
        #f.close()


    def read_and_write_data(self, basename, ddn, gdf_name):
        proc_names = glob(self.source_dir+'id*')
        #print 'Reading a dataset from %i Processor Files' % len(proc_names)
        N = len(proc_names)
        for i in range(N):
            if i == 0:
                fn = self.source_dir+'id%i/'%i + basename + '.%04i'%ddn + '.vtk'
            else:
                fn = self.source_dir+'id%i/'%i + basename + '-id%i'%i + '.%04i'%ddn + '.vtk'
            f = open(fn,'rb')
            #print 'Reading data from %s' % fn
            line = f.readline()
            while line != '':
                if len(line) == 0: break
                splitup = line.strip().split()

                if "DIMENSIONS" in splitup:
                    grid_dims = np.array(splitup[-3:]).astype('int')
                    line = f.readline()
                    continue
                elif "CELL_DATA" in splitup:
                    grid_ncells = int(splitup[-1])
                    line = f.readline()
                    if np.prod(grid_dims) != grid_ncells:
                        grid_dims -= 1
                        grid_dims[grid_dims==0]=1
                    if np.prod(grid_dims) != grid_ncells:
                        print('product of dimensions %i not equal to number of cells %i' % \
                              (np.prod(grid_dims), grid_ncells))
                        raise TypeError
                    break
                else:
                    del line
                    line = f.readline()
            read_table = False
            while line != '':
                if len(line) == 0: break
                splitup = line.strip().split()
                if 'SCALARS' in splitup:
                    field = splitup[1]
                    if not read_table:
                        line = f.readline() # Read the lookup table line
                        read_table = True
                    data = np.fromfile(f, dtype='>f4', count=grid_ncells).reshape(grid_dims,order='F')
                    if i == 0:
                        self.fields.append(field)
                    # print 'writing field %s' % field
                    self.write_gdf_field(gdf_name, i, field, data)
                    read_table=False

                elif 'VECTORS' in splitup:
                    field = splitup[1]
                    data = np.fromfile(f, dtype='>f4', count=3*grid_ncells)
                    data_x = data[0::3].reshape(grid_dims,order='F')
                    data_y = data[1::3].reshape(grid_dims,order='F')
                    data_z = data[2::3].reshape(grid_dims,order='F')
                    if i == 0:
                        self.fields.append(field+'_x')
                        self.fields.append(field+'_y')
                        self.fields.append(field+'_z')

                    # print 'writing field %s' % field
                    self.write_gdf_field(gdf_name, i, field+'_x', data_x)
                    self.write_gdf_field(gdf_name, i, field+'_y', data_y)
                    self.write_gdf_field(gdf_name, i, field+'_z', data_z)
                    del data, data_x, data_y, data_z
                del line
                line = f.readline()  # NOQA
            f.close()
            del f

        f = self.handle 
        field_g = f['field_types']
        # Add Field Attributes
        for name in self.fields:
            tname = name
            try:
                tname = translation_dict[name]
            except Exception: pass
            this_field = field_g.create_group(tname)
            if name in self.field_conversions.keys():
                this_field.attrs['field_to_cgs'] = self.field_conversions[name]
            else:
                this_field.attrs['field_to_cgs'] = np.float64('1.0') # For Now
            

    def convert(self, index=True, data=True):
        self.handle = h5.File(self.outname, 'a')
        if index:
            self.read_and_write_index(self.basename, self.ddn ,self.outname)
        if data:
            self.read_and_write_data(self.basename, self.ddn ,self.outname)
        self.handle.close()

class AthenaConverter(Converter):
    def __init__(self, basename, outname=None, field_conversions=None):
        self.fields = []
        self.basename = basename
        name = basename.split('.')
        fn = '%s.%04i'%(name[0],int(name[1]))
        self.ddn = int(name[1])
        self.basename = fn
        if outname is None:
            outname = fn+'.gdf'
        self.outname = outname
        if field_conversions is None:
            field_conversions = {}
        self.field_conversions = field_conversions


    def parse_line(self, line, grid):
        # grid is a dictionary
        splitup = line.strip().split()
        if "vtk" in splitup:
            grid['vtk_version'] = splitup[-1]
        elif "Really" in splitup:
            grid['time'] = splitup[-1]
        elif "DIMENSIONS" in splitup:
            grid['dimensions'] = np.array(splitup[-3:]).astype('int')
        elif "ORIGIN" in splitup:
            grid['left_edge'] = np.array(splitup[-3:]).astype('float64')
        elif "SPACING" in splitup:
            grid['dds'] = np.array(splitup[-3:]).astype('float64')
        elif "CELL_DATA" in splitup:
            grid["ncells"] = int(splitup[-1])
        elif "SCALARS" in splitup:
            field = splitup[1]
            grid['read_field'] = field
            grid['read_type'] = 'scalar'
        elif "VECTORS" in splitup:
            field = splitup[1]
            grid['read_field'] = field
            grid['read_type'] = 'vector'
        
    def read_grid(self, filename):
        """ Read Athena legacy vtk file from single cpu """
        f = open(filename,'rb')
        #print 'Reading from %s'%filename
        grid = {}
        grid['read_field'] = None
        grid['read_type'] = None
        table_read=False
        line = f.readline()
        while line != '':
            while grid['read_field'] is None:
                self.parse_line(line, grid)
                if grid['read_type'] == 'vector':
                    break
                if table_read is False:             
                    line = f.readline()
                if 'TABLE' in line.strip().split():
                    table_read = True
                if len(line) == 0: break

            if len(line) == 0: break
            if np.prod(grid['dimensions']) != grid['ncells']:
                grid['dimensions'] -= 1
            if np.prod(grid['dimensions']) != grid['ncells']:
                print('product of dimensions %i not equal to number of cells %i' % \
                      (np.prod(grid['dimensions']), grid['ncells']))
                raise TypeError

            if grid['read_type'] == 'scalar':
                grid[grid['read_field']] = \
                    np.fromfile(f, dtype='>f4', count=grid['ncells']).reshape(grid['dimensions'],order='F')
                self.fields.append(grid['read_field'])
            elif grid['read_type'] == 'vector':
                data = np.fromfile(f, dtype='>f4', count=3*grid['ncells'])
                grid[grid['read_field']+'_x'] = data[0::3].reshape(grid['dimensions'],order='F')
                grid[grid['read_field']+'_y'] = data[1::3].reshape(grid['dimensions'],order='F')
                grid[grid['read_field']+'_z'] = data[2::3].reshape(grid['dimensions'],order='F')
                self.fields.append(grid['read_field']+'_x')
                self.fields.append(grid['read_field']+'_y')
                self.fields.append(grid['read_field']+'_z')
            else:
                raise TypeError
            grid['read_field'] = None
            grid['read_type'] = None
            line = f.readline()
            if len(line) == 0: break
        grid['right_edge'] = grid['left_edge']+grid['dds']*(grid['dimensions'])
        return grid

    def write_to_gdf(self, fn, grid):
        f = h5.File(fn,'a')

        ## --------- Begin level nodes --------- ##
        g = f.create_group('gridded_data_format')
        g.attrs['format_version']=np.float32(1.0)
        g.attrs['data_software']='athena'
        data_g = f.create_group('data')
        field_g = f.create_group('field_types')
        f.create_group('particle_types')
        pars_g = f.create_group('simulation_parameters')

        dle = grid['left_edge'] # True only in this case of one grid for the domain
        gles = np.array([grid['left_edge']])
        gdims = np.array([grid['dimensions']])
        glis = ((gles - dle)/grid['dds']).astype('int64')

        # grid_left_index
        f.create_dataset('grid_left_index',data=glis)
        # grid_dimensions
        f.create_dataset('grid_dimensions',data=gdims)

        levels = np.array([0]).astype('int64') # unigrid example
        # grid_level
        f.create_dataset('grid_level',data=levels)

        ## ----------QUESTIONABLE NEXT LINE--------- ##
        # This data needs two dimensions for now. 
        n_particles = np.array([[0]]).astype('int64')
        #grid_particle_count
        f.create_dataset('grid_particle_count',data=n_particles)

        # Assume -1 means no parent.
        parent_ids = np.array([-1]).astype('int64')
        # grid_parent_id
        f.create_dataset('grid_parent_id',data=parent_ids)

        ## --------- Done with top level nodes --------- ##

        f.create_group('index')

        ## --------- Store Grid Data --------- ##

        g0 = data_g.create_group('grid_%010i'%0)
        for field in self.fields:
            name = field
            if field in translation_dict.keys():
                name = translation_dict[name]
            if name not in g0.keys(): 
                g0.create_dataset(name,data=grid[field])

        ## --------- Store Particle Data --------- ##

        # Nothing to do

        ## --------- Attribute Tables --------- ##

        pars_g.attrs['refine_by'] = np.int64(1)
        pars_g.attrs['dimensionality'] = np.int64(3)
        pars_g.attrs['domain_dimensions'] = grid['dimensions']
        try:
            pars_g.attrs['current_time'] = grid['time']
        except Exception:
            pars_g.attrs['current_time'] = 0.0
        pars_g.attrs['domain_left_edge'] = grid['left_edge'] # For Now
        pars_g.attrs['domain_right_edge'] = grid['right_edge'] # For Now
        pars_g.attrs['unique_identifier'] = 'athenatest'
        pars_g.attrs['cosmological_simulation'] = np.int64(0)
        pars_g.attrs['num_ghost_zones'] = np.int64(0)
        pars_g.attrs['field_ordering'] = np.int64(0)
        pars_g.attrs['boundary_conditions'] = np.int64([0]*6) # For Now

        # Extra pars:
        pars_g.attrs['n_cells'] = grid['ncells']
        pars_g.attrs['vtk_version'] = grid['vtk_version']

        # Add Field Attributes
        for name in g0.keys():
            tname = name
            try:
                tname = translation_dict[name]
            except Exception: pass
            this_field = field_g.create_group(tname)
        if name in self.field_conversions.keys():
            this_field.attrs['field_to_cgs'] = self.field_conversions[name]
        else:
            this_field.attrs['field_to_cgs'] = np.float64('1.0') # For Now

        # Add particle types
        # Nothing to do here

        # Add particle field attributes
        f.close()

    def convert(self):
        grid = self.read_grid(self.basename+'.vtk')
        self.write_to_gdf(self.outname,grid)
        
# import sys
# if __name__ == '__main__':
#     n = sys.argv[-1]
#     n = n.split('.')
#     fn = '%s.%04i'%(n[0],int(n[1]))
#     grid = read_grid(fn+'.vtk')
#     write_to_hdf5(fn+'.gdf',grid)
    



"""
The field detector.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

import numpy as np
from collections import defaultdict
from yt.units.yt_array import YTArray
from .field_exceptions import \
    NeedsGridType

class FieldDetector(defaultdict):
    Level = 1
    NumberOfParticles = 1
    _read_exception = None
    _id_offset = 0
    domain_id = 0

    def __init__(self, nd = 16, ds = None, flat = False, field_parameters=None):
        self.nd = nd
        self.flat = flat
        self._spatial = not flat
        self.ActiveDimensions = [nd,nd,nd]
        self.shape = tuple(self.ActiveDimensions)
        self.size = np.prod(self.ActiveDimensions)
        self.LeftEdge = [0.0, 0.0, 0.0]
        self.RightEdge = [1.0, 1.0, 1.0]
        self.dds = np.ones(3, "float64")
        if field_parameters is None:
            self.field_parameters = {}
        else:
            self.field_parameters = field_parameters
        class fake_dataset(defaultdict):
            pass

        if ds is None:
            # required attrs
            ds = fake_dataset(lambda: 1)
            ds["Massarr"] = np.ones(6)
            ds.current_redshift = ds.omega_lambda = ds.omega_matter = \
                ds.cosmological_simulation = 0.0
            ds.gamma = 5./3.0
            ds.hubble_constant = 0.7
            ds.domain_left_edge = np.zeros(3, 'float64')
            ds.domain_right_edge = np.ones(3, 'float64')
            ds.dimensionality = 3
            ds.periodicity = (True, True, True)
        self.ds = ds

        class fake_index(object):
            class fake_io(object):
                def _read_data_set(io_self, data, field):
                    return self._read_data(field)
                _read_exception = RuntimeError
            io = fake_io()
            def get_smallest_dx(self):
                return 1.0

        self.index = fake_index()
        self.requested = []
        self.requested_parameters = []
        if not self.flat:
            defaultdict.__init__(self,
                lambda: np.ones((nd, nd, nd), dtype='float64')
                + 1e-4*np.random.random((nd, nd, nd)))
        else:
            defaultdict.__init__(self,
                lambda: np.ones((nd * nd * nd), dtype='float64')
                + 1e-4*np.random.random((nd * nd * nd)))

    def _reshape_vals(self, arr):
        if not self._spatial: return arr
        if len(arr.shape) == 3: return arr
        return arr.reshape(self.ActiveDimensions, order="C")

    def __missing__(self, item):
        if not isinstance(item, tuple):
            field = ("unknown", item)
        else:
            field = item
        finfo = self.ds._get_field_info(*field)
        params, permute_params = finfo._get_needed_parameters(self)
        self.field_parameters.update(params)
        # For those cases where we are guessing the field type, we will
        # need to re-update -- otherwise, our item will always not have the
        # field type.  This can lead to, for instance, "unknown" particle
        # types not getting correctly identified.
        # Note that the *only* way this works is if we also fix our field
        # dependencies during checking.  Bug #627 talks about this.
        item = self.ds._last_freq
        if finfo is not None and finfo._function.__name__ != 'NullFunc':
            try:
                for param, param_v in permute_params.items():
                    for v in param_v:
                        self.field_parameters[param] = v
                        vv = finfo(self)
                if not permute_params:
                    vv = finfo(self)
            except NeedsGridType as exc:
                ngz = exc.ghost_zones
                nfd = FieldDetector(
                    self.nd + ngz * 2, ds = self.ds,
                    field_parameters=self.field_parameters.copy())
                nfd._num_ghost_zones = ngz
                vv = finfo(nfd)
                if ngz > 0: vv = vv[ngz:-ngz, ngz:-ngz, ngz:-ngz]
                for i in nfd.requested:
                    if i not in self.requested: self.requested.append(i)
                for i in nfd.requested_parameters:
                    if i not in self.requested_parameters:
                        self.requested_parameters.append(i)
            if vv is not None:
                if not self.flat: self[item] = vv
                else: self[item] = vv.ravel()
                return self[item]
        elif finfo is not None and finfo.particle_type:
            if "particle_position" in (item, item[1]) or \
               "particle_velocity" in (item, item[1]) or \
               "particle_magnetic_field" in (item, item[1]) or \
               "Velocity" in (item, item[1]) or \
               "Velocities" in (item, item[1]) or \
               "Coordinates" in (item, item[1]) or \
               "MagneticField" in (item, item[1]):
                # A vector
                self[item] = \
                  YTArray(np.ones((self.NumberOfParticles, 3)),
                          finfo.units, registry=self.ds.unit_registry)
            elif "FourMetalFractions" in (item, item[1]):
                self[item] = \
                  YTArray(np.ones((self.NumberOfParticles, 4)),
                          finfo.units, registry=self.ds.unit_registry)
            else:
                # Not a vector
                self[item] = \
                  YTArray(np.ones(self.NumberOfParticles),
                          finfo.units, registry=self.ds.unit_registry)
            if item == ('STAR', 'BIRTH_TIME'):
                # hack for the artio frontend so we pass valid times to
                # the artio functions for calculating physical times
                # from internal times
                self[item] *= -0.1
            self.requested.append(item)
            return self[item]
        self.requested.append(item)
        if item not in self:
            self[item] = self._read_data(item)
        return self[item]

    def _debug(self):
        # We allow this to pass through.
        return

    def deposit(self, *args, **kwargs):
        return np.random.random((self.nd, self.nd, self.nd))

    def mesh_sampling_particle_field(self, *args, **kwargs):
        pos = args[0]
        npart = len(pos)
        return np.random.rand(npart)

    def smooth(self, *args, **kwargs):
        tr = np.random.random((self.nd, self.nd, self.nd))
        if kwargs['method'] == "volume_weighted":
            return [tr]

        return tr

    def particle_operation(self, *args, **kwargs):
        return None

    def _read_data(self, field_name):
        self.requested.append(field_name)
        finfo = self.ds._get_field_info(*field_name)
        if finfo.particle_type:
            self.requested.append(field_name)
            return np.ones(self.NumberOfParticles)
        return YTArray(defaultdict.__missing__(self, field_name),
                       input_units=finfo.units,
                       registry=self.ds.unit_registry)

    fp_units = {
        'bulk_velocity' : 'cm/s',
        'bulk_magnetic_field': 'G',
        'center' : 'cm',
        'normal' : '',
        'cp_x_vec': '',
        'cp_y_vec': '',
        'cp_z_vec': '',
        'x_hat': '',
        'y_hat': '',
        'z_hat': '',
        'omega_baryon': '',
        'virial_radius': 'cm',
        'observer_redshift': '',
        'source_redshift': '',
        }

    def get_field_parameter(self, param, default = 0.0):
        if self.field_parameters and param in self.field_parameters:
            return self.field_parameters[param]
        self.requested_parameters.append(param)
        if param in ['center', 'normal'] or param.startswith('bulk'):
            return self.ds.arr(
                np.random.random(3) * 1e-2, self.fp_units[param])
        elif param in ['surface_height']:
            return self.ds.quan(0.0, 'code_length')
        elif param in ['axis']:
            return 0
        elif param.startswith("cp_"):
            ax = param[3]
            rv = self.ds.arr((0.0, 0.0, 0.0), self.fp_units[param])
            rv['xyz'.index(ax)] = 1.0
            return rv
        elif param.endswith("_hat"):
            ax = param[0]
            rv = YTArray((0.0, 0.0, 0.0), self.fp_units[param])
            rv['xyz'.index(ax)] = 1.0
            return rv
        elif param == "fof_groups":
            return None
        elif param == "mu":
            return 1.0
        else:
            return default

    _num_ghost_zones = 0
    id = 1

    def apply_units(self, arr, units):
        return self.ds.arr(arr, input_units = units)

    def has_field_parameter(self, param):
        return param in self.field_parameters

    @property
    def fcoords(self):
        fc = np.array(np.mgrid[0:1:self.nd*1j,
                               0:1:self.nd*1j,
                               0:1:self.nd*1j])
        if self.flat:
            fc.shape = (self.nd*self.nd*self.nd, 3)
        else:
            fc = fc.transpose()
        return self.ds.arr(fc, input_units = "code_length")

    @property
    def fcoords_vertex(self):
        fc = np.random.random((self.nd, self.nd, self.nd, 8, 3))
        if self.flat:
            fc.shape = (self.nd*self.nd*self.nd, 8, 3)
        return self.ds.arr(fc, input_units = "code_length")

    @property
    def icoords(self):
        ic = np.mgrid[0:self.nd-1:self.nd*1j,
                      0:self.nd-1:self.nd*1j,
                      0:self.nd-1:self.nd*1j]
        if self.flat:
            ic.shape = (self.nd*self.nd*self.nd, 3)
        else:
            ic = ic.transpose()
        return ic

    @property
    def ires(self):
        ir = np.ones(self.nd**3, dtype="int64")
        if not self.flat:
            ir.shape = (self.nd, self.nd, self.nd)
        return ir

    @property
    def fwidth(self):
        fw = np.ones((self.nd**3, 3), dtype="float64") / self.nd
        if not self.flat:
            fw.shape = (self.nd, self.nd, self.nd, 3)
        return self.ds.arr(fw, input_units = "code_length")




"""MOAB-specific fields


"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from yt.fields.field_info_container import \
    FieldInfoContainer

class MoabFieldInfo(FieldInfoContainer):
    pass

class PyneFieldInfo(FieldInfoContainer):
    pass



"""
Very simple convenience function for importing all the modules, setting up
the namespace and getting the last argument on the command line.



"""

#-----------------------------------------------------------------------------
# Copyright (c) 2013, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

from __future__ import absolute_import

#
# ALL IMPORTS GO HERE
#

import os
from yt import *

# This next item will handle most of the actual startup procedures, but it will
# also attempt to parse the command line and set up the global state of various
# operations.  The variable unparsed_args is not used internally but is
# provided as a convenience for users who wish to parse arguments in scripts.
# https://mail.python.org/archives/list/yt-dev@python.org/thread/L6AQPJ3OIMJC5SNKVM7CJG32YVQZRJWA/
import yt.startup_tasks as __startup_tasks
unparsed_args = __startup_tasks.unparsed_args

from yt.config import ytcfg, ytcfg_defaults

from yt.utilities.logger import level as __level
if __level >= int(ytcfg_defaults["loglevel"]):
    # This won't get displayed.
    mylog.debug("Turning off NumPy error reporting")
    np.seterr(all = 'ignore')

# We load plugins.  Keep in mind, this can be fairly dangerous -
# the primary purpose is to allow people to have a set of functions
# that get used every time that they don't have to *define* every time.
# This way, other command-line tools can be used very simply.
# Unfortunately, for now, I think the easiest and simplest way of doing
# this is also the most dangerous way.
if ytcfg.getboolean("yt","loadfieldplugins"):
    enable_plugins()



import struct
import numpy as np

# Size of basic types (in bytes)
SIZE_LOGICAL = 4
SIZE_INT = 4
SIZE_DOUBLE = 8
NAME_LEN = 16

# For un-aligned data, use '=' (for aligned data set to '')
ALIGN = '='

def get_header(istream):
    """Read header from an MPI-AMRVAC 2.0 snapshot.
    istream' should be a file
    opened in binary mode.
    """
    istream.seek(0)
    h = {}

    fmt = ALIGN + 'i'
    [h['datfile_version']] = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))

    if h['datfile_version'] < 3:
        raise IOError("Unsupported AMRVAC .dat file version: %d", h['datfile_version'])

    # Read scalar data at beginning of file
    fmt = ALIGN + 9 * 'i' + 'd'
    hdr = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))
    [h['offset_tree'], h['offset_blocks'], h['nw'],
     h['ndir'], h['ndim'], h['levmax'], h['nleafs'], h['nparents'],
     h['it'], h['time']] = hdr

    # Read min/max coordinates
    fmt = ALIGN + h['ndim'] * 'd'
    h['xmin'] = np.array(
        struct.unpack(fmt, istream.read(struct.calcsize(fmt))))
    h['xmax'] = np.array(
        struct.unpack(fmt, istream.read(struct.calcsize(fmt))))

    # Read domain and block size (in number of cells)
    fmt = ALIGN + h['ndim'] * 'i'
    h['domain_nx'] = np.array(
        struct.unpack(fmt, istream.read(struct.calcsize(fmt))))
    h['block_nx'] = np.array(
        struct.unpack(fmt, istream.read(struct.calcsize(fmt))))

    if h['datfile_version'] >= 5:
        # Read periodicity
        fmt = ALIGN + h['ndim'] * 'i' # Fortran logical is 4 byte int
        h['periodic'] = np.array(
            struct.unpack(fmt, istream.read(struct.calcsize(fmt))), dtype=bool)

        # Read geometry name
        fmt = ALIGN + NAME_LEN * 'c'
        hdr = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))
        h['geometry'] = b''.join(hdr).strip().decode()

        # Read staggered flag
        fmt = ALIGN + 'i' # Fortran logical is 4 byte int
        h['staggered'] = bool(
            struct.unpack(fmt, istream.read(struct.calcsize(fmt)))[0])

    # Read w_names
    w_names = []
    for i in range(h['nw']):
        fmt = ALIGN + NAME_LEN * 'c'
        hdr = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))
        w_names.append(b''.join(hdr).strip().decode())
    h['w_names'] = w_names

    # Read physics type
    fmt = ALIGN + NAME_LEN * 'c'
    hdr = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))
    h['physics_type'] = b''.join(hdr).strip().decode()

    # Read number of physics-defined parameters
    fmt = ALIGN + 'i'
    [n_pars] = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))

    # First physics-parameter values are given, then their names
    fmt = ALIGN + n_pars * 'd'
    vals = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))

    fmt = ALIGN + n_pars * NAME_LEN * 'c'
    names = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))
    # Split and join the name strings (from one character array)
    names = [b''.join(names[i:i+NAME_LEN]).strip().decode()
             for i in range(0, len(names), NAME_LEN)]

    # Store the values corresponding to the names
    for val, name in zip(vals, names):
        h[name] = val
    return h

def get_tree_info(istream):
    """
    Read levels, morton-curve indices, and byte offsets for each block as stored in the datfile
    istream is an open datfile buffer with 'rb' mode
    This can be used as the "first pass" data reading required by YT's interface.
    """
    istream.seek(0)
    header = get_header(istream)
    nleafs = header['nleafs']
    nparents = header['nparents']

    # Read tree info. Skip 'leaf' array
    istream.seek(header['offset_tree'] + (nleafs+nparents) * SIZE_LOGICAL)

    # Read block levels
    fmt = ALIGN + nleafs * 'i'
    block_lvls = np.array(struct.unpack(fmt, istream.read(struct.calcsize(fmt))))

    # Read block indices
    fmt = ALIGN + nleafs * header['ndim'] * 'i'
    block_ixs = np.reshape(struct.unpack(fmt, istream.read(struct.calcsize(fmt))),
                           [nleafs, header['ndim']])

    # Read block offsets (skip ghost cells !)
    bcfmt = ALIGN + header['ndim'] * 'i'
    bcsize = struct.calcsize(bcfmt) * 2

    fmt = ALIGN + nleafs * 'q'
    block_offsets = np.array(struct.unpack(fmt, istream.read(struct.calcsize(fmt)))) + bcsize
    return block_lvls, block_ixs, block_offsets

def get_single_block_data(istream, byte_offset, block_shape):
    """retrieve a specific block (all fields) from a datfile"""
    istream.seek(byte_offset)
    # Read actual data
    fmt = ALIGN + np.prod(block_shape) * 'd'
    d = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))
    # Fortran ordering
    block_data = np.reshape(d, block_shape, order='F')
    return block_data


def get_single_block_field_data(istream, byte_offset, block_shape, field_idx):
    """retrieve a specific block (ONE field) from a datfile"""
    istream.seek(byte_offset)

    # compute byte size of a single field
    field_shape = block_shape[:-1]
    fmt = ALIGN + np.prod(field_shape) * 'd'
    byte_size_field = struct.calcsize(fmt)

    # Read actual data
    istream.seek(byte_size_field * field_idx, 1) # seek forward
    d = struct.unpack(fmt, istream.read(struct.calcsize(fmt)))

    # Fortran ordering
    block_field_data = np.reshape(d, field_shape, order='F')
    return block_field_data



"""
A volume container




"""

#-----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

cimport numpy as np

cdef struct VolumeContainer:
    #-----------------------------------------------------------------------------
    # Encapsulates a volume container used for volume rendering.
    #
    #    n_fields       int              : The number of fields available to the volume renderer.
    #    data           np.float64_t**   : The data within the volume container.
    #    mask           np.uint8_t*      : The mask of the volume container. It has dimensions one fewer in each direction than data.       
    #    left_edge      np.float64_t[3]  : The left edge of the volume container's bounding box.
    #    right_edge     np.float64_t[3]  : The right edge of the volume container's bounding box.
    #    np.float64_t   dds[3]           : The delta dimensions, such that dds[0] = ddx, dds[1] = ddy, dds[2] = ddz.
    #    np.float64_t   idds[3]          : The inverse delta dimensions. Same as dds pattern, but the inverse. i.e. idds[0] = inv_ddx.
    #    dims           int[3]           : The dimensions of the volume container. dims[0] = x, dims[1] = y, dims[2] = z.
    #-----------------------------------------------------------------------------
    int n_fields
    np.float64_t **data
    np.uint8_t *mask
    np.float64_t left_edge[3]
    np.float64_t right_edge[3]
    np.float64_t dds[3]
    np.float64_t idds[3]
    int dims[3]

cdef inline int vc_index(VolumeContainer *vc, int i, int j, int k):
    #-----------------------------------------------------------------------------
    # vc_index(VolumeContainer *vc, int i, int j, int k)
    #    vc   VolumeContainer* : Pointer to the volume container to be indexed.
    #    i    int              : The first dimension coordinate.
    #    j    int              : The second dimension coordinate.
    #    k    int              : The third dimension coordinates.
    #
    # Returns the 3-dimensional index in the volume container given coordinates i, j, k.
    # This is used for 3-dimensional access in a flat container using C-ordering.
    # This is calculated by: 
    #       vc_index = i * vc.dim[1] * vc.dims[2] + j * vc.dims[2] + k
    # and then simplified (as shown below) by combining one multiplication operation.
    #
    # 2-dimensional example:
    #       A 4 x 3 array may be represented as: 
    #                                      a = [0,  1,  2,  3,
    #                                           4,  5,  6,  7,
    #                                           8,  9,  10, 11]
    #       or similarly, in a flat container in row-successive order as:
    #                          a = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
    #
    # To access the coordinate at (1,1) in the flat container:
    #                         i * dims[1] + j
    #                       = 1 *   3     + 1
    #                       = 4
    # The 3-dimensional case (presented below) is similar.
    #-----------------------------------------------------------------------------
    return (i*vc.dims[1]+j)*vc.dims[2]+k

cdef inline int vc_pos_index(VolumeContainer *vc, np.float64_t *spos):
    cdef int index[3]
    cdef int i
    for i in range(3):
        index[i] = <int> ((spos[i] - vc.left_edge[i]) * vc.idds[i])
    return vc_index(vc, index[0], index[1], index[2])
